#!/usr/bin/env python

"""
Usage: converter.py <dataset dir> <save dir> <input width> <input height> <train set size>
    dataset dir: path to dataset root directory
    save dir: path to converted data output directory
    max image size (eg. 140 140)
    train subset relative size (eg. 0.6)
"""
import os
import sys
import random
from datasets.imdb_wiki_face_dataset import IMDB_WIKI_FACE
import h5py
import math
from neon.backends import gen_backend


def shuffle(arr):
    result = []
    for _ in range(len(arr)):
        source_index = random.randint(0, len(arr) - 1)
        result.append(arr.pop(source_index))
    return result

def compute_stats(dataset_iterator):
    gender_0_count = 0
    gender_1_count = 0
    for i, e in dataset_iterator:
        if int(e[1].get()) == 0:
            gender_0_count += 1
        else:
            gender_1_count += 1
    print("Class balance: gender = 0: %d, gender = 1: %d" \
        % (gender_0_count, gender_1_count))

if (__name__ == '__main__'):
    if (len(sys.argv) != 6):
        print(__doc__)
        exit(0)

    dataset_root = os.path.abspath(sys.argv[1])
    save_dir = os.path.abspath(sys.argv[2])
    max_image_size = (int(sys.argv[3]), int(sys.argv[4]))
    train_set_size = float(sys.argv[5])
    if ((train_set_size <= 0) or (1 < train_set_size)):
        print("Expected train subset size in range [0, 1]")
        exit(1)
    random.seed(42)

    dataset = IMDB_WIKI_FACE(dataset_root, save_dir)
    dataset_metadata = dataset.get_metadata()

    # filter dataset
    dataset_genders = \
        dataset_metadata[IMDB_WIKI_FACE.MetadataCategories.GENDER][0]
    dataset_indices = []
    for i, gender in enumerate(dataset_genders):
        if (math.isnan(gender) == True):
            continue
        dataset_indices.append(i)

    # make train/validation set separation
    dataset_indices = shuffle(dataset_indices)
    subset_boundary = int(len(dataset_indices) * train_set_size) + 1
    train_indices = dataset_indices[:subset_boundary]
    val_indices = dataset_indices[subset_boundary:]

    dataset.prepare_data(train_indices, val_indices, save_dir, max_image_size)

    gen_backend('cpu', batch_size=1)
    train_it, val_it = dataset.gen_iterators()
    compute_stats(train_it)
    compute_stats(val_it)

    # Example: accessing datatset entries
    # import neon.backends
    # import numpy as np
    # import PIL.Image

    # neon.backends.gen_backend('cpu', batch_size=1)
    # dataset = IMDB_WIKI_FACE(data_root='.')
    # train_set, val_set = dataset.gen_iterators()

    # train_it = iter(train_set)
    # entry = next(train_it) # on device memory
    # image_data = entry[0].get() # to host memory
    # gender_data = entry[1].get() # to host memory

    # image = PIL.Image.fromarray(image_data.astype('i1') \
    #    .reshape((3, 128, 128)).transpose((1, 2, 0)), 'RGB')
    # gender = gender_data[0]
