#!/usr/bin/env python

from neon.callbacks.callbacks import Callbacks
from neon.models import Model
from neon.optimizers import GradientDescentMomentum, MultiOptimizer
from neon.util.argparser import NeonArgparser

import datasets
from datasets.imdb_wiki_face_dataset import IMDB_WIKI_FACE
from datasets.iterators import HDF5Iterator

import models
from models.layers import MSE
from models import import_matching_layers

from time import time
import os
import h5py as h5
import numpy as np
from itertools import chain
from contextlib import closing



def parse_args(docstring):
    parser = NeonArgparser(docstring)
    parser.add_argument('--model_name', default=None)
    parser.add_argument('--data_root', default='./data')
    parser.add_argument('--restrict_dataset', default='0')
    return parser.parse_args()


def create_model(model_name, args={}):
    model_generator = models.get_model_generator(model_name)
    if (model_generator is None):
        print("Not found model '%s'" % (model_name))
        exit(1)
    print("Creating model '" + model_name + "'")
    return model_generator(**args)


def get_training_steps_list(model_name):
    # Get list of training steps for stacked autoencoder:
    return create_model(model_name + "_steps")


def generate_dataset(subset_name, subset, outputs, save_path):
    inputs = iter(subset)
    sample_input = next(iter(subset))[0].get().transpose()[0]
    lshape = sample_input.shape
    input_size = int(np.prod(lshape))
    output_size = int(np.prod(outputs[0].shape))
    with closing(h5.File(save_path, 'w')) as dataset_file:
        dataset_inputs = dataset_file.create_dataset('input',
            (len(outputs), input_size), dtype=np.int8)
        dataset_inputs.attrs['lshape'] = lshape
        dataset_outputs = dataset_file.create_dataset('output',
            (len(outputs), output_size), dtype=np.int8)
        i = 0
        for batch in inputs:
            for inp in batch[0].get().transpose():
                if (len(outputs) <= i):
                    break
                output = outputs[i]
                dataset_inputs[i] = inp.reshape((-1)).astype(np.int8)
                dataset_outputs[i] = output.reshape((-1)).astype(np.int8)
                i += 1


def generate_next_dataset(model, model_name, train_set, val_set, save_path, step):
    # Drop decoder part and generate outputs
    lshape = next(iter(train_set))[0].get()[:, 0].shape
    encoder, _ = create_model(model_name + '_encoder', {'input_shape': lshape})
    import_matching_layers(model.get_description(get_weights=True), encoder)

    save_path_base = save_path[:save_path.rfind(".")] + "_outputs_step" + str(step)

    def _make_subset(subset_name, subset):
        outputs = encoder.get_outputs(subset)
        subset_path = save_path_base + "_" + subset_name + ".h5"
        generate_dataset(subset_name, subset, outputs, subset_path)
        return HDF5Iterator(subset_path), outputs[0].shape
    next_train_set, lshape = _make_subset('train', train_set)
    next_val_set, _ = _make_subset('val', val_set)

    return next_train_set, next_val_set, lshape


if (__name__ == '__main__'):
    args = parse_args(__doc__)
    data_root = os.path.abspath(args.data_root)
    restrict_dataset = int(args.restrict_dataset)
    model_name = args.model_name

    model_steps = get_training_steps_list(model_name)
    print("Trining steps:", model_steps)

    # Data loading:
    dataset = IMDB_WIKI_FACE(data_root=data_root)
    train_set, val_set, lshape = dataset.gen_iterators_autoencoder()

    prev_model = None
    for step, step_model_name in enumerate(model_steps):
        model, cost = create_model(step_model_name, {'input_shape': lshape})

        optimizer = None
        if prev_model is not None:
            imported_layers = import_matching_layers(
                prev_model.get_description(get_weights=True), model)

            # Make optimizer for transfer learning:
            default_optimizer = GradientDescentMomentum(
                0.1, momentum_coef=0.9, stochastic_round=args.rounding, wdecay=0.0005)
            optimizers_mapping = {'default': default_optimizer}
            # Make optimizers with small learning rate for imported layers
            imported_layers_optimizer_params = \
                default_optimizer.get_description().copy()['config']
            imported_layers_optimizer_params['learning_rate'] = 0
            imported_layers_optimizer = \
                GradientDescentMomentum(**imported_layers_optimizer_params)
            optimizers_mapping.update({l: imported_layers_optimizer for l in imported_layers})
            optimizer = MultiOptimizer(optimizers_mapping)
        else:
            optimizer = GradientDescentMomentum(
                0.1, momentum_coef=0.9, stochastic_round=args.rounding, wdecay=0.0005)

        callbacks = Callbacks(model, eval_set=val_set, **args.callback_args)
        fit_time = time()
        model.fit(train_set, optimizer=optimizer,
            num_epochs=args.epochs, cost=cost, callbacks=callbacks)
        fit_time = time() - fit_time

        if args.save_path is not None:
            model.save_params(args.save_path + ".step_" + str(step))

        print('Step %d: Learning time: %ds' % (step, fit_time))
        inference_time = time()
        accuracy = model.eval(val_set, metric=MSE())
        inference_time = time() - inference_time
        print('Step %d: MSE = %.1f' % (step, accuracy))
        print('Step %d: Inference time: %ds' % (step, inference_time))

        train_set, val_set, lshape = \
            generate_next_dataset(model, step_model_name, train_set, val_set, 
                args.save_path, step)

        prev_model = model
    final_ae_model = prev_model

    if args.save_path is not None:
        final_ae_model.save_params(args.save_path)
