#!/usr/bin/env python

"""
Parameters:
    --source_model <filepath>
    --target_model <filepath>
    --target_model_name <name>
    --data_root=<path> [default: data]
    --restrict_dataset=<n> [default: 1000]
    and all of the Neon default parameters
"""

from neon.callbacks.callbacks import Callbacks
from neon.optimizers import GradientDescentMomentum, MultiOptimizer
from neon.transforms import Accuracy
from neon.models import Model
from neon.backends import gen_backend
from neon.util.persist import load_obj

import datasets
from datasets.imdb_wiki_face_dataset import IMDB_WIKI_FACE
from datasets.dummy import generate_data

import models
from models import import_matching_layers

import os
from time import time
import numpy as np


def save_inference(output_file_name, outputs, subset):
    with open(output_file_name, 'w') as output_file:
        output_file.write('inference, target\n')
        outputs_iter = iter(outputs)
        try:
            for dataset_batch in subset:
                targets = np.transpose(dataset_batch[1].get())
                for target in targets:
                    output = next(outputs_iter)
                    target_class = np.argmax(target, axis=0)
                    output_file.write('%s, %s\n' % (output, target_class))
        except StopIteration as e: # the last batch might be incomplete
            pass

def parse_args(docstring):
    from neon.util.argparser import NeonArgparser
    parser = NeonArgparser(docstring)
    parser.add_argument('--source_model')
    parser.add_argument('--target_model', default=None)
    parser.add_argument('--target_model_name')
    parser.add_argument('--data_root', default='./data')
    parser.add_argument('--restrict_dataset', default='0')
    return parser.parse_args()

if (__name__ == '__main__'):
    args = parse_args(__doc__)
    pretrained_model_filepath = args.source_model
    target_model_filepath = args.target_model
    target_model_name = args.target_model_name
    data_root = os.path.abspath(args.data_root)
    restrict_dataset = int(args.restrict_dataset)

    # Preparing datasets
    dataset = IMDB_WIKI_FACE(data_root=data_root)
    train_set, val_set, lshape = dataset.gen_iterators()

    # Create the model
    target_model_generator = models.get_model_generator(target_model_name)
    if (target_model_generator is None):
        print("Not found model '%s'" % (target_model_name))
        exit(1)
    target_model, cost = target_model_generator()
    # target_model.initialize(train_set) # needed to perform weights and shapes initialization in the model

    pretrained_model = load_obj(pretrained_model_filepath)
    imported_layers = import_matching_layers(pretrained_model, target_model)

    # Make optimizer for transfer learning:
    base_lr = 0.001
    default_optimizer = GradientDescentMomentum(
        base_lr, momentum_coef=0.9, stochastic_round=args.rounding, wdecay=0.0005)
    optimizers_mapping = {'default': default_optimizer}
    # Make optimizers with small learning rates for imported layers
    imported_layers_optimizer_params = \
        default_optimizer.get_description().copy()['config']
    imported_layers_optimizer_params['learning_rate'] = 0 # base_lr # * 0.0001
    imported_layers_optimizer = \
        GradientDescentMomentum(**imported_layers_optimizer_params)
    optimizers_mapping.update({l: imported_layers_optimizer for l in imported_layers})
    optimizer = MultiOptimizer(optimizers_mapping)


    # Fit the model
    callbacks = Callbacks(target_model, eval_set=val_set, **args.callback_args)
    fit_time = time()
    target_model.fit(train_set, optimizer=optimizer,
        num_epochs=args.epochs, cost=cost, callbacks=callbacks)
    fit_time = time() - fit_time
    print('Learning time: %ds' % (fit_time))
    inference_time = time()
    accuracy = target_model.eval(val_set, metric=Accuracy())
    inference_time = time() - inference_time
    print('Accuracy = %.1f%%' % (accuracy * 100))
    print('Inference time: %ds' % (inference_time))

    if (target_model_filepath is not None):
        target_model.save_params(target_model_filepath)

    print('Making inference')
    outputs = target_model.get_outputs(val_set)
    save_inference('inference.txt', outputs, val_set)