#!/usr/bin/env python

from neon.callbacks.callbacks import Callbacks
from neon.models import Model
from neon.optimizers import GradientDescentMomentum
from neon.util.argparser import NeonArgparser
from neon.transforms import Accuracy

import datasets
from datasets.iterators import ArrayIterator
from datasets.imdb_wiki_face_dataset import IMDB_WIKI_FACE
from datasets.dummy import generate_data

import models
from models import get_weights_norm

from time import time
import os
import numpy as np


def save_inference(output_file_name, outputs, subset):
    with open(output_file_name, 'w') as output_file:
        output_file.write('inference, target\n')
        outputs_iter = iter(outputs)
        try:
            for dataset_batch in subset:
                targets = np.transpose(dataset_batch[1].get())
                for target in targets:
                    output = next(outputs_iter)
                    target_class = np.argmax(target, axis=0)
                    output_file.write('%s, %s\n' % (output, target_class))
        except StopIteration as e: # the last batch might be incomplete
            pass

def print_activations(inputs, model):
    print("Inputs:\n", inputs.get())
    for i, l in enumerate(model.layers.layers):
        inputs = l.fprop(inputs, True)
        print("Layer %d\n%s:\n%s" % (i, l, inputs.get()))

def parse_args(docstring):
    from neon.util.argparser import NeonArgparser
    parser = NeonArgparser(docstring)
    parser.add_argument('--model_name', default=None)
    parser.add_argument('--data_root', default='./data')
    parser.add_argument('--restrict_dataset', default='0')
    return parser.parse_args()

if (__name__ == '__main__'):
    args = parse_args(__doc__)
    data_root = os.path.abspath(args.data_root)
    restrict_dataset = int(args.restrict_dataset)
    model_name = args.model_name

    dataset = IMDB_WIKI_FACE(data_root=data_root)
    # Data loading in 3 variants:
    # 1. Load only current batch every time
    train_set, val_set, lshape = dataset.gen_iterators()
    
    # 2. Load all dataset once and hold it in the device memory
    # train_set, val_set, lshape = dataset.gen_iterators_inmemory()

    # 3. Same as 2, but hand-made
    # X_train, y_train, X_val, y_val, lshape = \
    #     dataset.load(restrict_dataset, restrict_dataset)
    # X_train, y_train, X_val, y_val, lshape = generate_data((3, 128, 128), args.batch_size)
    # train_set = ArrayIterator(X=X_train, y=y_train, nclass=2, lshape=lshape)
    # val_set = ArrayIterator(X=X_val, y=y_val, nclass=2, lshape=lshape)

    model_generator = None
    if (model_name is not None):
        model_generator = models.get_model_generator(model_name)
        if (model_generator is None):
            print("Not found model '%s'" % (model_name))
            exit(1)
        print("Using model '" + model_name + "'")
    else:
        model_generator = models.generate_mlp3_model
    model, cost = model_generator()
    # Load model with: --model_file path

    optimizer = GradientDescentMomentum(
        0.01, momentum_coef=0.9, stochastic_round=args.rounding, wdecay=0.0005)

    callbacks = Callbacks(model, eval_set=val_set, **args.callback_args)
    
    fit_time = time()
    model.fit(train_set, optimizer=optimizer,
        num_epochs=args.epochs, cost=cost, callbacks=callbacks)
    fit_time = time() - fit_time
    print('Learning time: %ds' % (fit_time))
    
    inference_time = time()
    accuracy = model.eval(val_set, metric=Accuracy())
    inference_time = time() - inference_time
    print('Accuracy = %.1f%%' % (accuracy * 100))
    print('Inference time: %ds' % (inference_time))

    # Save training state of the model: --serialize iterations_step --save_path path

    print("Train norm:", get_weights_norm(model))

    print('Making inference')
    outputs = model.get_outputs(val_set)
    save_inference('inference.txt', outputs, val_set)

    # Print activations layer by layer:
    # inputs_iter = iter(train_set)
    # inputs = next(inputs_iter)[0]
    # print_activations(inputs, model)
