#!/usr/bin/env python

from neon.callbacks.callbacks import Callbacks
from neon.models import Model
from neon.optimizers import GradientDescentMomentum, PowerSchedule
from neon.util.argparser import NeonArgparser
from neon.transforms import Accuracy
from neon.util.persist import load_obj

import datasets
from datasets.camvid import CamVid

import models
from models import get_weights_norm
from models.layers import ConfusionMetric

from time import time
import os
import numpy as np


def parse_args(docstring):
    from neon.util.argparser import NeonArgparser
    parser = NeonArgparser(docstring)
    parser.add_argument('--model_name', default=None)
    parser.add_argument('--data_root', default='./data_camvid')
    parser.add_argument('--restrict_dataset', default='0')
    parser.add_argument('--inference_path', default=None)
    parser.add_argument('--source_model', default=None)
    return parser.parse_args()

def print_activations(inputs, model):
    print("Inputs:\n", inputs.get())
    for i, l in enumerate(model.layers.layers):
        inputs = l.fprop(inputs, True)
        view = inputs.get().reshape(l.out_shape)
        print("Layer %d\n%s:\n%s" % (i, l, view))

def save_inference(output_dir, outputs, filename_list, dataset, class_count, width, height):
    import PIL.Image as Image

    os.makedirs(output_dir, exist_ok=True)

    for i, output_raw in enumerate(outputs):
        output = output_raw.reshape((class_count, height, width))
        output_grayscale = Image.fromarray(np.argmax(output, axis=0).astype(np.int8), mode='L')
        output_painted = dataset.paint_segmentation(output_grayscale)
        file_name = filename_list[i][0]
        file_name = file_name[:file_name.rfind('.')]
        output_painted.save(output_dir + '/' + file_name + '.png')

if (__name__ == '__main__'):
    args = parse_args(__doc__)
    data_root = os.path.abspath(args.data_root)
    restrict_dataset = int(args.restrict_dataset)
    model_name = args.model_name
    inference_path = args.inference_path

    dataset = CamVid(data_root=data_root)
    train_set, val_set, lshape = dataset.gen_iterators()
    # train_set, val_set, lshape = dataset.gen_iterators_inmemory(restrict_dataset, restrict_dataset)

    model_generator = None
    if (model_name is not None):
        model_generator = models.get_model_generator(model_name)
        if (model_generator is None):
            print("Not found model '%s'" % (model_name))
            exit(1)
        print("Using model '" + model_name + "'")
    else:
        model_generator = models.generate_cnn1_seg_model
    model, cost = model_generator()
    # Load model with: --model_file path

    if (args.source_model is not None):
        pretrained_model = load_obj(args.source_model)
        models.import_matching_layers(pretrained_model, model)

    schedule = PowerSchedule(200, 0.1)
    optimizer = GradientDescentMomentum(
        0.00001, momentum_coef=0.9, stochastic_round=args.rounding, wdecay=0.0005,
        schedule=schedule)

    callbacks = Callbacks(model, eval_set=val_set, **args.callback_args)
    
    fit_time = time()
    model.fit(train_set, optimizer=optimizer,
        num_epochs=args.epochs, cost=cost, callbacks=callbacks)
    fit_time = time() - fit_time
    print('Learning time: %ds' % (fit_time))
    print("Train norm:", get_weights_norm(model))
    
    metric = ConfusionMetric(dataset.classes)
    inference_time = time()
    accuracy = model.eval(val_set, metric=metric)
    inference_time = time() - inference_time
    print('Model score:')
    metric.show_results()
    print('Inference time: %ds' % (inference_time))

    # Save training state of the model: --serialize iterations_step --save_path path

    # inputs_iter = iter(train_set)
    # inputs = next(inputs_iter)[0]
    # print_activations(inputs, model)

    if (inference_path is not None):
        print('Saving inference to \'%s\'' % inference_path)
        outputs = model.get_outputs(val_set)
        height, width = lshape[1:]
        save_inference(os.path.abspath(inference_path), outputs, dataset.get_lists()[1], \
            dataset, dataset.class_count, width=width, height=height)
