import os
import h5py as h5
import numpy as np
from PIL import Image
from datasets.dataset_base import BasicDataset
from datasets.iterators import HDF5Iterator, HDF5IteratorOneHot, HDF5IteratorAutoencoder, ArrayIterator, PixelwiseIterator


"""
Class color mapping:
00. 0 0 0 Void
01. 192 0 128 Archway
02. 0 128 192 Bicyclist
03. 0 128 64 Bridge
04. 128 0 0  Building
05. 64 0 128 Car
06. 64 0 192 CartLuggagePram
07. 192 128 64 Child
08. 192 192 128 Column_Pole
09. 64 64 128 Fence
10. 128 0 192 LaneMkgsDriv
11. 192 0 64 LaneMkgsNonDriv
12. 128 128 64 Misc_Text
13. 192 0 192 MotorcycleScooter
14. 128 64 64 OtherMoving
15. 64 192 128 ParkingBlock
16. 64 64 0 Pedestrian
17. 128 64 128 Road
18. 128 128 192 RoadShoulder
19. 0 0 192 Sidewalk
20. 192 128 128 SignSymbol
21. 128 128 128 Sky
22. 64 128 192 SUVPickupTruck
23. 0 0 64 TrafficCone
24. 0 64 64 TrafficLight
25. 192 64 128 Train
26. 128 128 0 Tree
27. 192 128 192 Truck_Bus
28. 64 0 64 Tunnel
29. 192 192 0 VegetationMisc
30. 64 128 64 Animal
31. 64 192 0 Wall
"""

def class_colormap_grayscale_to_index(p):
    if p == 0: return 0
    if p == 72: return 1
    if p == 97: return 2
    if p == 82: return 3
    if p == 38: return 4
    if p == 33: return 5
    if p == 41: return 6
    if p == 139: return 7
    if p == 184: return 8
    if p == 71: return 9
    if p == 60: return 10
    if p == 64: return 11
    if p == 120: return 12
    if p == 79: return 13
    if p == 83: return 14
    if p == 146: return 15
    if p == 56: return 16
    if p == 90: return 17
    if p == 135: return 18
    if p == 21: return 19
    if p == 147: return 20
    if p == 128: return 21
    if p == 116: return 22
    if p == 7: return 23
    if p == 44: return 24
    if p == 109: return 25
    if p == 113: return 26
    if p == 154: return 27
    if p == 26: return 28
    if p == 170: return 29
    if p == 101: return 30
    if p == 131: return 31
    if p == 255: return 0
    return 0

class_colormap_index_to_rgb_table = [
    0, 0, 0,
    192, 0, 128,
    0, 128, 192,
    0, 128, 64,
    128, 0, 0,
    64, 0, 128,
    64, 0, 192,
    192, 128, 64,
    192, 192, 128,
    64, 64, 128,
    128, 0, 192,
    192, 0, 64,
    128, 128, 64,
    192, 0, 192,
    128, 64, 64,
    64, 192, 128,
    64, 64, 0,
    128, 64, 128,
    128, 128, 192,
    0, 0, 192,
    192, 128, 128,
    128, 128, 128,
    64, 128, 192,
    0, 0, 64,
    0, 64, 64,
    192, 64, 128,
    128, 128, 0,
    192, 128, 192,
    64, 0, 64,
    192, 192, 0,
    64, 128, 64,
    64, 192, 0,
]

def class_colormap_index_to_r(p):
    if (32 <= p): return 255
    return class_colormap_index_to_rgb_table[p * 3 + 0]

def class_colormap_index_to_g(p):
    if (32 <= p): return 255
    return class_colormap_index_to_rgb_table[p * 3 + 1]

def class_colormap_index_to_b(p):
    if (32 <= p): return 255
    return class_colormap_index_to_rgb_table[p * 3 + 2]

class CamVid(BasicDataset):
    SEQUENCES = [
        # (source file, output format string)
        ('01TP_extract.avi', '0001TP_%06d_L.jpg'),
        ('0005VD.MXF', 'Seq05VD_f%06d_L.jpg'),
        ('0006R0.MXF', '0006R0_f%06d_L.jpg'),
        ('0016E5.MXF', '0016E5_%06d_L.jpg')
    ]
    class_count = 32
    classes = [
        'Void',
        'Archway',
        'Bicyclist',
        'Bridge',
        'Building',
        'Car',
        'CartLuggagePram',
        'Child',
        'Column_Pole',
        'Fence',
        'LaneMkgsDriv',
        'LaneMkgsNonDriv',
        'Misc_Text',
        'MotorcycleScooter',
        'OtherMoving',
        'ParkingBlock',
        'Pedestrian',
        'Road',
        'RoadShoulder',
        'Sidewalk',
        'SignSymbol',
        'Sky',
        'SUVPickupTruck',
        'TrafficCone',
        'TrafficLight',
        'Train',
        'Tree',
        'Truck_Bus',
        'Tunnel',
        'VegetationMisc',
        'Animal',
        'Wall'
    ]

    def __init__(self, dataset_root = '.', data_root = '.',
            train_set_file = BasicDataset.default_train_set_file_name,
            val_set_file = BasicDataset.default_val_set_file_name):
        super().__init__(dataset_root, data_root, train_set_file, val_set_file)
        self.inputs_dir = 'images'
        self.outputs_dir = 'gt_images'

    def _adopt_grayscale_gt(self, gt_image):
        return gt_image.point(class_colormap_grayscale_to_index)

    def paint_segmentation(self, image):
        painted_channels = [None] * 3
        painted_channels[0] = image.point(class_colormap_index_to_r)
        painted_channels[1] = image.point(class_colormap_index_to_g)
        painted_channels[2] = image.point(class_colormap_index_to_b)
        return Image.merge("RGB", painted_channels)

    def _load_subset(self, subset_data, lshape, subset_max_size=0):
        subset_size = len(subset_data['input'])
        if (0 < subset_max_size):
            subset_size = min(subset_size, subset_max_size)
        subset_mean = subset_data['input'].attrs['mean']
        lshape_flatten = int(np.prod(lshape))
        X = np.empty((subset_size, lshape_flatten), dtype=float)
        y = np.empty((subset_size, lshape[1] * lshape[2]), dtype=int)
        for i, entry in enumerate(zip(subset_data['input'], subset_data['output'])):
            if (i >= subset_size):
                break
            if (i % max(1, subset_size // 10) == 0):
                print("\rLoading dataset... (%d / %d)" % (i, subset_size - 1), end='\r')

            X[i] = (entry[0].reshape(lshape).transpose((1, 2, 0)) - subset_mean). \
                transpose((2, 0, 1)).flatten()
            y[i] = entry[1]
        print()

        return (X, y)

    def _make_dataset(self, filepath, images, image_size):
        inputs_dir = self.dataset_root + '/' + self.inputs_dir
        outputs_dir = self.dataset_root + '/' + self.outputs_dir

        dataset_file = h5.File(filepath + '.h5', 'w')
        dataset_list = open(filepath + '_list.txt', 'w')

        dataset_size = len(images)
        input_channels_count = 3
        input_channel_size = image_size[0] * image_size[1]
        input_size = input_channels_count * input_channel_size
        output_size = input_channel_size

        dataset_inputs = dataset_file.create_dataset('input',
            (dataset_size, input_size), dtype=np.int8)
        dataset_inputs.attrs['lshape'] = \
            (input_channels_count, image_size[1], image_size[0])
        dataset_outputs = dataset_file.create_dataset('output',
            (dataset_size, output_size), dtype=np.int8)

        channels_mean = np.zeros((input_channels_count, ))
        for pos, entry in enumerate(images):
            if (pos % 100 == 0):
                print("\rConverting... (%d / %d)" % (pos, dataset_size - 1), end='\r')

            input_image = Image.open(inputs_dir + "/" + entry[0]).convert('RGB')
            input_image = self._make_center_crop(input_image, image_size)
            image_array = np.array(input_image, dtype=np.int8).transpose((2, 0, 1))
            channels_mean += np.mean(image_array, axis=(1, 2))
            dataset_inputs[pos] = image_array.flatten()

            output_image = Image.open(outputs_dir + "/" + entry[1])
            output_image = self._make_center_crop(output_image, image_size)
            if (len(output_image.getbands()) == 3):
                # Should be done only if segmentation is in RGB format
                output_image = self._adopt_grayscale_gt(output_image.convert('L'))
            elif (len(output_image.getbands()) != 1):
                raise Exception('Unknown segmentation format. Only indexed and RGB (painted) are accepted.')
            output_image = np.array(output_image, dtype=np.int8)
            dataset_outputs[pos] = output_image.flatten()

            dataset_list.write(entry[0] + ', ' + entry[1] + '\n')
        print()

        dataset_inputs.attrs['mean'] = channels_mean / dataset_size

        dataset_file.close()
        dataset_list.close()

    def extract_frames(self):
        save_dir = self.dataset_root + '/images'
        if (os.path.exists(save_dir) == False):
            os.makedirs(save_dir)

        for seq in self.SEQUENCES:
            subprocess.check_result([
                'ffmpeg',
                '-i', self.dataset_root + '/' + seq[0],
                '-framerate', '30',
                save_dir + '/' + seq[1]])

    def create_list(self):
        dataset_entries = []
        input_list = os.listdir(self.dataset_root + '/' + self.inputs_dir)
        output_list = os.listdir(self.dataset_root + '/' + self.outputs_dir)
        for output_name in output_list:
            if output_name.find('.png') == -1:
                continue
            input_name = output_name[:-4] + ".jpg"
            if input_name in input_list:
                dataset_entries.append((input_name, output_name))

        return dataset_entries

    def prepare_data(self, train_list, val_list, save_dir, input_size):
        """
        train_list, val_list - iterables with string pairs
        save_dir - path to dataset files output directory
        input_size = (width, height)

        Prepares dataset in HDF5 format for later use.
        """

        if (os.path.exists(save_dir) == False):
            os.makedirs(save_dir)

        self._make_dataset(save_dir + '/' + self.train_set_file_name,
            train_list, input_size)
        self._make_dataset(save_dir + '/' + self.val_set_file_name,
            val_list, input_size)

    def gen_iterators(self):
        train_set_path = \
            self.data_root + '/' + self.train_set_file_name + '.h5'
        val_set_path = \
            self.data_root + '/' + self.val_set_file_name + '.h5'
        lshape = None
        with h5.File(train_set_path, 'r') as f:
            lshape = tuple([int(v) for v in f['input'].attrs['lshape']])
            f.close()
        train_set = HDF5Iterator(train_set_path)
        val_set = HDF5Iterator(val_set_path)
        train_set = PixelwiseIterator(train_set, self.class_count)
        val_set = PixelwiseIterator(val_set, self.class_count)
        return (train_set, val_set, lshape)

    def gen_iterators_inmemory(self, restrict_train=0, restrict_val=0):
        X_train, y_train, X_val, y_val, lshape = self.load(restrict_train, restrict_val)
        train_set = ArrayIterator(X=X_train, y=y_train, lshape=lshape, make_onehot=False)
        val_set = ArrayIterator(X=X_val, y=y_val, lshape=lshape, make_onehot=False)
        train_set = PixelwiseIterator(train_set, self.class_count)
        val_set = PixelwiseIterator(val_set, self.class_count)
        return (train_set, val_set, lshape)

    def get_lists(self):
        train_set_list = [
            line.split(' ') for line \
            in open(self.data_root + '/' + self.train_set_file_name + '_list.txt')
        ]
        val_set_list = [
            line.split(' ') for line \
            in open(self.data_root + '/' + self.val_set_file_name + '_list.txt')
        ]
        return train_set_list, val_set_list
