import os
from scipy.io import loadmat
import h5py as h5
import numpy as np
import math
from PIL import Image
from datasets.dataset_base import BasicDataset
from datasets.iterators import HDF5Iterator, HDF5IteratorOneHot, HDF5IteratorAutoencoder, ArrayIterator


class IMDB_WIKI_FACE(BasicDataset):
    class MetadataCategories:
        BIRTH_DATE = 0
        PHOTO_DATE = 1
        IMAGE_PATH = 2
        GENDER = 3
        NAME = 4
        FACE_LOCATION = 5
        FACE_SCORE = 6
        SECOND_FACE_SCORE = 7

    def __init__(self, dataset_root = '.', data_root = '.', 
            train_set_file = BasicDataset.default_train_set_file_name, 
            val_set_file = BasicDataset.default_val_set_file_name):
        super().__init__(dataset_root, data_root, train_set_file, val_set_file)

    def _load_subset(self, subset_data, lshape, subset_max_size=0):
        subset_size = len(subset_data['input'])
        if (0 < subset_max_size):
            subset_size = min(subset_size, subset_max_size)
        subset_mean = subset_data['input'].attrs['mean']
        lshape_flatten = int(np.prod(lshape))
        X = np.empty((subset_size, lshape_flatten), dtype=float)
        y = np.empty((subset_size, 1), dtype=int)
        for i, entry in enumerate(zip(subset_data['input'], subset_data['output'])):
            if (i >= subset_size):
                break
            if (i % (subset_size // 10) == 0):
                print("\rLoading dataset... (%d / %d)" % (i, subset_size - 1), end='\r')

            X[i] = (entry[0].reshape(lshape).transpose((1, 2, 0)) - subset_mean). \
                transpose((2, 0, 1)).flatten()
            y[i] = entry[1]
        print()

        return (X, y)

    def _make_checked_indices_list(self, unchecked_list, metadata):
        checked_list = list()
        for i in unchecked_list:
            gender = metadata[self.MetadataCategories.GENDER][0][i]
            if (math.isnan(gender) == False):
                checked_list.append(i)
            else:
                print("Gender for entry #" + str(i) + \
                    " is not defined. Skipping this entry.")
        return checked_list

    def _make_dataset(self, filepath, indices, metadata, image_size):
        dataset_file = h5.File(filepath + '.h5', 'w')
        dataset_list = open(filepath + '_list.txt', 'w')
        indices = self._make_checked_indices_list(indices, metadata)

        input_channels_count = 3
        input_channel_size = image_size[0] * image_size[1]
        input_size = input_channels_count * input_channel_size

        dataset_inputs = dataset_file.create_dataset('input', 
            (len(indices), input_size), dtype=np.int8)
        dataset_inputs.attrs['lshape'] = \
            (input_channels_count, image_size[1], image_size[0])
        dataset_outputs = dataset_file.create_dataset('output', 
            (len(indices), 1), dtype=int)
        dataset_outputs.attrs['nclass'] = 2

        channels_mean = np.zeros((input_channels_count, ))
        for pos, i in enumerate(indices):
            if (pos % 100 == 0):
                print("\rConverting... (%d / %d)" % (pos, len(indices) - 1), end='\r')

            gender = metadata[self.MetadataCategories.GENDER][0][i].astype(int)

            image_path = self.dataset_root + '/' + \
                metadata[self.MetadataCategories.IMAGE_PATH][0][i][0]
            image = Image.open(image_path)
            image = self._make_center_crop(image, image_size)
            if (len(image.getbands()) != 3):
                image = image.convert("RGB")

            image_array = np.array(image, dtype=np.int8).transpose((2, 0, 1))
            channels_mean += np.mean(image_array, axis=(1, 2))
            dataset_inputs[pos] = image_array.flatten()
            dataset_outputs[pos] = gender

            dataset_list.write(image_path + '\n')
        print()

        dataset_inputs.attrs['mean'] = channels_mean / len(indices)

        dataset_file.close()
        dataset_list.close()

    def prepare_data(self, train_indices, val_indices, save_dir, input_size):
        """
        train_indices, val_indices - iterable with integer values
        save_dir - path to dataset files output directory
        input_size = (width, height) [default: None]

        Prepares dataset in HDF5 format for later use.
        """

        metadata = self.get_metadata()

        if (os.path.exists(save_dir) == False):
            os.makedirs(save_dir)

        self._make_dataset(save_dir + '/' + self.train_set_file_name, 
            train_indices, metadata, input_size)
        self._make_dataset(save_dir + '/' + self.val_set_file_name, 
            val_indices, metadata, input_size)

    def get_metadata(self, metadata_file = 'wiki'):
        meta_file = loadmat(self.dataset_root + '/' + metadata_file)
        return meta_file['wiki'][0][0]

    def gen_iterators(self):
        train_set_path = \
            self.data_root + '/' + self.train_set_file_name + '.h5'
        val_set_path = \
            self.data_root + '/' + self.val_set_file_name + '.h5'
        lshape = None
        with h5.File(train_set_path, 'r') as f:
            lshape = tuple([int(v) for v in f['input'].attrs['lshape']])
            f.close()
        train_set_iterator = HDF5IteratorOneHot(train_set_path)
        val_set_iterator = HDF5IteratorOneHot(val_set_path)
        return (train_set_iterator, val_set_iterator, lshape)

    def gen_iterators_autoencoder(self):
        train_set_path = \
            self.data_root + '/' + self.train_set_file_name + '.h5'
        val_set_path = \
            self.data_root + '/' + self.val_set_file_name + '.h5'
        lshape = None
        with h5.File(train_set_path, 'r') as f:
            lshape = tuple([int(v) for v in f['input'].attrs['lshape']])
            f.close()
        train_set_iterator = HDF5IteratorAutoencoder(train_set_path)
        val_set_iterator = HDF5IteratorAutoencoder(val_set_path)
        return (train_set_iterator, val_set_iterator, lshape)

    def gen_iterators_inmemory(self):
        X_train, y_train, X_val, y_val, lshape = self.load()
        train_set = ArrayIterator(X=X_train, y=y_train, nclass=2, lshape=lshape)
        val_set = ArrayIterator(X=X_val, y=y_val, nclass=2, lshape=lshape)
        return (train_set, val_set, lshape)

    def gen_iterators_autoencoder_inmemory(self):
        X_train, y_train, X_val, y_val, lshape = self.load()
        train_set = ArrayIterator(X_train, lshape=lshape)
        val_set = ArrayIterator(X_val, lshape=lshape)
        return (train_set, val_set, lshape)
