# ----------------------------------------------------------------------------
# Copyright 2014 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
"""
Defines basic input datatset types.
"""
import os
import h5py
import logging
import numpy as np
from neon import NervanaObject
from neon.data import NervanaDataIterator
from builtins import zip


class ArrayIterator(NervanaDataIterator):

    """
    The ArrayIterator class iterates over minibatches of data that
    have been preloaded into memory in the form of numpy arrays. This may be used when
    the entire dataset (e.g. CIFAR-10 or MNIST) is small enough to fit in memory. For example::

        X = np.random.rand(10000, 3072)
        y = np.random.randint(0, 10, 10000)
        train = ArrayIterator(X=X, y=y, nclass=10, lshape=(3, 32, 32))

    The above will create the ArrayIterator object. This object implements python's __iter__
    method, and returns one minibatch of data, formatted as tuple of (input, label), with
    each iteration. The minibatch size is controlled by the generated backend.

    X should be an ndarray of shape (# example, # features). For images, the features should be
    formatted in (channel, height, width) order. The `lshape` keyword indicates the local shape of
    the images in (channel, height, width) format.

    For classification tasks, the labels `y` should be integers from 0 to K-1, where K is the total
    number of classes. When `y` is not provided, the input features themselves will be returned
    as the target values (e.g. autoencoder).

    In regression tasks, where `y` is not a categorical label, set `make_onehot` to `False`.
    For example::

        X = np.random.rand(1000, 1)
        y = 2*X + 1
        train = ArrayIterator(X=X, y=y, make_onehot=False)

    For more information, see the Loading data section of the documentation.
    """

    def __init__(self, X, y=None, nclass=None, lshape=None, make_onehot=True, name=None):
        """
        During initialization, the input data will be converted to backend tensor objects
        (e.g. CPUTensor or GPUTensor). If the backend uses the GPU, the data is copied over to the
        device.

        Args:
            X (ndarray, shape: [# examples, feature size]): Input features of the
                dataset.
            y (ndarray, shape:[# examples, 1 or feature size], optional): Labels corresponding to
                the input features. If absent, the input features themselves will be returned as
                target values (e.g. autoencoder)
            nclass (int, optional): The number of classes in labels. Not necessary if
                labels are not provided or where the labels are non-categorical.
            lshape (tuple, optional): Local shape for the input features
                (e.g. # channels, height, width)
            make_onehot (bool, optional): True if y is a categorical label that has to be converted
                to a one hot representation.

        """
        # Treat singletons like list so that iteration follows same syntax
        super(ArrayIterator, self).__init__(name=name)
        X = X if isinstance(X, list) else [X]
        self.ndata = len(X[0])
        assert self.ndata >= self.be.bsz
        self.start = 0
        self.nclass = nclass
        self.ybuf = None

        if make_onehot and nclass is None and y is not None:
            raise AttributeError('Must provide number of classes when creating onehot labels')

        # if labels provided, they must have same # examples as the features
        if y is not None:

            assert all([y.shape[0] == x.shape[0] for x in X]), \
                "Input features and labels must have equal number of examples."

            # for classifiction, the labels must be from 0 .. K-1, where K=nclass
            if make_onehot:
                assert y.max() <= nclass - 1 and y.min() >= 0, \
                    "Labels must range from 0 to {} (nclass-1).".format(nclass - 1)

                assert (np.floor(y) == y).all(), \
                    "Labels must only contain integers."

        # if local shape is provided, then the product of lshape should match the
        # number of features
        if lshape is not None:
            assert all([x.shape[1] == np.prod(lshape) for x in X]), \
                "product of lshape {} does not match input feature size".format(lshape)

        # store shape of the input data
        self.shape = [x.shape[1] if lshape is None else lshape for x in X]
        if len(self.shape) == 1:
            self.shape = self.shape[0]
            self.lshape = lshape

        # Helpers to make dataset, minibatch, unpacking function for transpose and onehot
        def transpose_gen(z):
            return (self.be.array(z), self.be.iobuf(z.shape[1]),
                    lambda _in, _out: self.be.copy_transpose(_in, _out))

        def onehot_gen(z):
            return (self.be.array(z.reshape((-1, 1)), dtype=np.int32), self.be.iobuf(nclass),
                    lambda _in, _out: self.be.onehot(_in, axis=0, out=_out))

        self.Xdev, self.Xbuf, self.unpack_func = list(zip(*[transpose_gen(x) for x in X]))

        # Shallow copies for appending, iterating
        self.dbuf, self.hbuf = list(self.Xdev), list(self.Xbuf)
        self.unpack_func = list(self.unpack_func)

        if y is not None:
            self.ydev, self.ybuf, yfunc = onehot_gen(y) if make_onehot else transpose_gen(y)
            self.dbuf.append(self.ydev)
            self.hbuf.append(self.ybuf)
            self.unpack_func.append(yfunc)

    @property
    def nbatches(self):
        """
        Return the number of minibatches in this dataset.
        """
        return -((self.start - self.ndata) // self.be.bsz)

    def reset(self):
        """
        Resets the starting index of this dataset to zero. Useful for calling
        repeated evaluations on the dataset without having to wrap around
        the last uneven minibatch. Not necessary when data is divisible by batch size
        """
        self.start = 0

    def __iter__(self):
        """
        Returns a new minibatch of data with each call.

        Yields:
            tuple: The next minibatch which includes both features and labels.
        """
        for i1 in range(self.start, self.ndata, self.be.bsz):
            bsz = min(self.be.bsz, self.ndata - i1)
            islice1, oslice1 = slice(0, bsz), slice(i1, i1 + bsz)
            islice2, oslice2 = None, None
            if self.be.bsz > bsz:
                islice2, oslice2 = slice(bsz, None), slice(0, self.be.bsz - bsz)
                self.start = self.be.bsz - bsz

            for buf, dev, unpack_func in zip(self.hbuf, self.dbuf, self.unpack_func):
                unpack_func(dev[oslice1], buf[:, islice1])
                if oslice2:
                    unpack_func(dev[oslice2], buf[:, islice2])

            inputs = self.Xbuf[0] if len(self.Xbuf) == 1 else self.Xbuf
            targets = self.ybuf if self.ybuf else inputs
            yield (inputs, targets)


class HDF5Iterator(ArrayIterator):
    """
    Data iterator which uses an HDF5 file as the source of the data, useful when
    the entire HDF5 dataset cannot fit into memory (for smaller datasets, use the ArrayIterator).

    To initialize the HDF5Iterator, simply call::

        train_set = HDF5Iterator('your_data_path.h5')

    The HDF5 file format must contain the following datasets:

    - `input` (ndarray):
                        Input data, which is a 2-D array (float or uint8) that
                        has size `(N, F)`, where `N` is the number of examples,
                        and `F` is the number of features. For images, `F = C*H*W` where
                        `C` is the number of channels and `H` and `W` are the height
                        and width of the image, respectively. This data must also
                        have the following attributes:
        - `lshape` (tuple):
                    Tuple of ints indicating the shape of each
                    input (for examples, image data may have
                    an lshape of `[C, H, W]`)
        - `mean` (ndarray):
                    The mean to subtract, either formatted as (F, 1) or
                    a mean for each channel with dimensions (C, 1)
    - `output` (ndarray):
                        An optional dataset which, if supplied, will be
                        used at the target/expected output of the network. the
                        array should have the shape `(N, M)` where `N` is the number
                        of items (must match the `N` dim of the input set)
                        and `M` is the size of the output data which must match
                        size of ouput from the output layer of the network.

    For cases where the output should be converted to a one-hot encoding (see Loading Data),
    use the `HDF5IteratorOneHot`. Or for autoencoder problems, use `HDFIteratorAutoencoder`.
    """
    def __init__(self, hdf_filename, name=None):
        """
        Args:
            hdf_filename (string): Path to the HDF5 datafile.
            name (string, optional): Name to assign this iterator. Defaults to None.
        """
        super(ArrayIterator, self).__init__(name=name)

        self.hdf_filename = hdf_filename

        if not os.path.isfile(hdf_filename):
            raise IOError('File not found %s' % hdf_filename)
        self.hdf_file = h5py.File(hdf_filename, mode='r', driver=None)

        # input data array
        self.inp = self.hdf_file['input']
        self.ndata = self.inp.shape[0]

        # must have at least 1 minibatch of data in the file
        assert self.ndata >= self.be.bsz
        self.start = 0

        # the input array unflattened size
        self.lshape = tuple([int(e) for e in self.inp.attrs['lshape']])
        self.shape = self.lshape

        if 'output' in self.hdf_file:
            self.out = self.hdf_file['output']

        self.inpbuf = None
        self.outbuf = None
        self.allocated = False

    def allocate(self):
        """
        After the input and output (`self.inp` and `self.out)` have been
        set this function will allocate the host and device buffers
        for the mini-batches.

        The host buffer is referenced as `self.mini_batch_in` and `self.mini_batch_out`, and
        stored on device as `self.inbuf` and `self.outbuf`.
        """
        if not self.allocated:
            self.allocate_inputs()
            self.allocate_outputs()
            self.allocated = True

    def allocate_inputs(self):
        """
        Allocates the host and device input data buffers
        and any other associated storage.

        `self.inpbuf` is the on-device buffer for the input minibatch
        `self.mini_batch_in` is the on-host buffer for the input minibatch
        `self.mean` is the on-device buffer of the mean array
        """
        # on device minibatch_buffer (input)
        self.inpbuf = self.be.iobuf(self.inp.shape[1])

        # setup host buffer for a mini_batch
        self.mini_batch_in = np.zeros(self.inpbuf.shape)

        self.mean = None
        # the 'mean' dataset is the the mean values to subtract
        if 'mean' in self.hdf_file:
            mns_ = np.array(self.hdf_file['mean']).flatten()
            if mns_.size != self.inp.shape[1]:
                # channel by channel mean
                # there should be 1 element per channel
                assert mns_.size == self.lshape[0], 'mean image size mismatch'
                # need to have 2-d array for broadcasting
                mns_ = mns_.reshape((self.lshape[0], 1)).copy()
                # make channel-by-channel mean subtraction view
                self.meansub_view = self.inpbuf.reshape((self.lshape[0], -1))
            else:
                self.meansub_view = self.inpbuf

            self.mean = self.be.array(mns_)

    def allocate_outputs(self):
        """
        Allocates the host and device output data buffers
        and any other associated storage.

        `self.outbuf` is the on-device buffer for the output minibatch
        `self.mini_batch_out` is the on-host buffer for the output minibatch
        """
        self.outbuf = None
        if 'output' in self.hdf_file:
            self.outbuf = self.be.iobuf(self.out.shape[1])
            self.mini_batch_out = np.zeros(self.outbuf.shape)

    def gen_input(self, mini_batch):
        """
        Function to handle any preprocessing before pushing an input
        mini-batch to the device.  For example, mean subtraction etc.

        Arguments:
            mini_batch (ndarray): M-by-N array where M is the flatten
                                  input vector size and N is the batch size
        """
        self.inpbuf[:] = mini_batch
        # mean subtract
        if self.mean is not None:
            self.meansub_view[:] = -self.mean + self.meansub_view

    def gen_output(self, mini_batch):
        """
        Function to handle any preprocessing before pushing an output
        mini-batch to the device.  For example, one-hot generation.

        Arguments:
            mini_batch (ndarray): M-by-N array where M is the flatten
                                  output vector size and N is the batch size
        """
        self.outbuf[:] = mini_batch

    def __del__(self):
        self.cleanup()
        # needed for python3 to exit cleanly
        super(HDF5Iterator, self).__del__()

    def cleanup(self):
        """
        Closes the HDF file.
        """
        self.hdf_file.close()

    def reset(self):
        """
        Resets the index to zero.
        """
        self.start = 0

    def __iter__(self):
        """
        Defines a generator that can be used to iterate over this dataset.

        Yields:
            tuple: The next minibatch. A minibatch includes both features and
            labels.
        """
        if not self.allocated:
            self.allocate()
        full_shape = list(self.lshape)
        full_shape.append(-1)

        mini_batch_in = self.mini_batch_in
        if self.outbuf is not None:
            mini_batch_out = self.mini_batch_out
        for i1 in range(self.start, self.ndata, self.be.bsz):
            i2 = min(i1 + self.be.bsz, self.ndata)
            bsz = i2 - i1
            if i2 == self.ndata:
                self.start = self.be.bsz - bsz

            # load mini batch on host
            xdev = self.inp
            mini_batch_in[:, :bsz] = xdev[i1:i2, :].T.astype(np.float32)
            if self.be.bsz > bsz:
                mini_batch_in[:, bsz:] = xdev[:(self.be.bsz - bsz), :].T.astype(np.float32)

            # push to device
            self.gen_input(mini_batch_in)

            if self.outbuf is not None:
                mini_batch_out[:, :bsz] = self.out[i1:i2].T
                if self.be.bsz > bsz:
                    mini_batch_out[:, bsz:] = self.out[:(self.be.bsz - bsz)].T

                self.gen_output(mini_batch_out)

            inputs = self.inpbuf
            targets = self.outbuf
            yield (inputs, targets)


class HDF5IteratorOneHot(HDF5Iterator):
    """
    Extends the HDF5Iterator class to add one hot conversion of the
    target data. For example::

        train_set = HDF5IteratorOneHot('your_data_path.h5')

    The HDF5 file format must contain the following datasets:
    - `input` dataset: Input data, which is a 2-D array (float or uint8) that
                        has size `(N, C*H*W)`, where `N` is the number of examples,
                        `C` is the number of channels, and `H` and `W` are the height
                        and width of the image, respectively. This dataset must also
                        have the following attributes:
        - `lshape`: a tuple of int indicating the shape of each
                    input (for examples, image data may have
                    an lshape of `[C, H, W]`)
        - `mean`: the mean to subtract, either formatted as (C*H*W, 1) or
                  a mean for each channel with dimensions (C, 1)
    - `output` dataset: An optional dataset which, if supplied, will be
                        used at the target/expected output of the network. the
                        array should have the shape `(N, M)` where `N` is the number
                        of items (must match the `N` dim of the input set)
                        and `M` is the size of the output data which must match
                        size of ouput from the output layer of the network.

    The "output" dataset in the HDF5 (if present) must have the 'nclass'
    attribute specifying the number of total output classes which is needed
    for generating the one-hot encoding.
    """
    def __init__(self, hdf_filename, name=None):
        """
        Args:
            hdf_filename (string): Path to the HDF5 datafile.
            name (string, optional): Name to assign this iterator. Defaults to None.
        """
        super(HDF5IteratorOneHot, self).__init__(hdf_filename, name=name)
        if 'output' in self.hdf_file:
            assert 'nclass' in self.hdf_file['output'].attrs, 'Missing nclass attribute'
            self.nclass = int(self.hdf_file['output'].attrs['nclass'])
            self.out = np.array(self.hdf_file['output'], dtype=np.int32).reshape((-1, 1))

    def allocate_outputs(self):
        self.outbuf = None
        if 'output' in self.hdf_file:
            self.argmax_buf = self.be.iobuf(1, dtype=np.int32)
            self.mini_batch_out = np.zeros(self.argmax_buf.shape, dtype=np.int32)
            self.outbuf = self.be.iobuf(self.nclass)

    def gen_output(self, mini_batch):
        self.argmax_buf[:] = mini_batch
        self.be.onehot(self.argmax_buf, axis=0, out=self.outbuf)


class HDF5IteratorAutoencoder(HDF5Iterator):
    """
    Extends the base HDF5Iterator class for an autoencoder model.
    Returns the input data as the target.
    """
    def __iter__(self):
        """
        Defines a generator that can be used to iterate over this dataset.

        Yields:
            tuple: The next minibatch containing the inputs and the targets (here target=inputs)
        """
        for x, t in super(HDF5IteratorAutoencoder, self).__iter__():
            yield (x, x)


class PixelwiseIterator:
    def __init__(self, target_iterator, nclass):
        self._be = NervanaObject().be
        self._target = target_iterator
        self._nclass = nclass

    def __iter__(self):
        for inputs, targets in self._target:
            targets = self._make_onehot(targets, self._nclass)
            yield (inputs, targets)

    def _make_onehot(self, targets, nclass):
        # This code is drawn from SegNet example:
        # https://github.com/NervanaSystems/neon_segnet/blob/master/pixelwise_dataloader.py

        t = targets
        # generate the extra buffers needed to hold the expanded
        # one-hot target output
        out_full = self._be.iobuf(nclass * t.shape[0], dtype=np.int32)
        outview = out_full.reshape((nclass, -1))

        # need an int32 copy of the target classes for one-hot to work correctly
        t_int32 = self._be.iobuf(t.shape[0], dtype=np.int32)
        t_int32[:] = t

        tview = t_int32.reshape((1, -1))
        # use backend one-hot to expand output to nclass x H x W
        self._be.onehot(tview, axis=0, out=outview)

        return out_full

    def __getattr__(self, attr):
        return getattr(self._target, attr)
