from neon.initializers import Gaussian, Kaiming, Constant, Xavier, GlorotUniform
from neon.layers import GeneralizedCost, Affine, Conv, Deconv, Pooling, BatchNorm, Activation, Dropout, DataTransform, Linear, Bias, MergeSum, SkipNode
from neon.transforms import Rectlin, Logistic, Softmax, PixelwiseSoftmax, Tanh, CrossEntropyBinary, CrossEntropyMulti, SumSquared, Normalizer, Explin
from neon.models import Model


def generate_cnn1_model():
    class_count = 2

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # convolutional encoder / feature extractor
        # resolution 1
        Conv(fshape=(3, 3, 32), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),

        # resolution 1/2
        Conv(fshape=(3, 3, 64), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),

        # classifier,
        Affine(nout=class_count, init=Gaussian(scale=0.1), bias=Constant(0), activation=Logistic(shortcut=True))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyBinary())

    return (model, cost)

def generate_cnn2_model():
    class_count = 2

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # receptive noise filter
        Conv(fshape=(3, 3, 32), padding=2, strides=1, dilation=2, init=Kaiming(), activation=Explin()),
        Pooling(fshape=(5, 5), padding=2, strides=2, op='max'),

        # convolutional encoder / feature extractor
        # resolution 1/2
        Conv(fshape=(3, 3, 32), padding=1, strides=1, dilation=1, init=Kaiming(), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, 64), padding=0, strides=2, dilation=1, init=Kaiming(), activation=Explin()),
        BatchNorm(),

        # resolution 1/4
        Conv(fshape=(3, 3, 64), padding=1, strides=1, dilation=1, init=Kaiming(), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, 128), padding=0, strides=2, dilation=1, init=Kaiming(), activation=Explin()),
        BatchNorm(),

        # resolution 1/8
        Conv(fshape=(3, 3, 128), padding=1, strides=1, dilation=1, init=Kaiming(), activation=Explin()),
        BatchNorm(),

        # classifier, global avg pooling from NIN'14
        Conv(fshape=(3, 3, class_count), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Pooling(fshape='all', padding=0, strides=1, op='avg'),
        Affine(nout=class_count, init=Gaussian(0.1), bias=Constant(0), activation=Logistic(shortcut=True))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyBinary())

    return (model, cost)

def generate_cnn3_model():
    class_count = 2

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # receptive noise filter
        Conv(fshape=(3, 3, 32), padding=2, strides=2, dilation=2, init=Kaiming(), bias=Constant(0), activation=Explin()),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),

        # convolutional encoder / feature extractor
        # resolution 1/4

        BatchNorm(),
        Conv(fshape=(1, 1, 32), padding=0, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(3, 3, 32), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, 64), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),

        # resolution 1/8

        BatchNorm(),
        Conv(fshape=(1, 1, 64), padding=0, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(3, 3, 64), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, 128), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),

        # resolution 1/16

        BatchNorm(),
        Conv(fshape=(1, 1, 128), padding=0, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(3, 3, 128), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, 256), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),

        # resolution 1/32

        # classifier, global avg pooling from NIN'14
        BatchNorm(),
        Conv(fshape=(3, 3, class_count), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        BatchNorm(),
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Explin()),
        Pooling(fshape='all', padding=0, strides=1, op='avg'),
        BatchNorm(),
        Affine(nout=class_count, init=Gaussian(0.1), bias=Constant(0), activation=Logistic(shortcut=True))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyBinary())

    return (model, cost)

def generate_tygrikon_cnn_model():
    # Adopted CNN model of https://github.com/Tygrikon/deep_learning_labs

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        Conv(fshape=(5, 5, 64), padding=2, strides=1, dilation=1, init=Xavier(), bias=Constant(0)),
        BatchNorm(),
        # Bias(init=Constant(0)),
        # Scale, Bias
        Activation(Rectlin()),
        Pooling(fshape=(2, 2), padding=0, strides=2, op='max'),

        Conv(fshape=(5, 5, 64), padding=2, strides=1, dilation=1, init=Xavier(), bias=Constant(0)),
        BatchNorm(),
        # Scale, Bias
        # Bias(init=Constant(0)),
        Activation(Rectlin()),
        Pooling(fshape=(2, 2), padding=0, strides=2, op='max'),

        Conv(fshape=(5, 5, 64), padding=2, strides=1, dilation=1, init=Xavier(), bias=Constant(0)),
        BatchNorm(),
        # Bias(init=Constant(0)),
        # Scale, Bias
        Activation(Rectlin()),
        Pooling(fshape=(2, 2), padding=0, strides=2, op='max'),

        Dropout(),
        Affine(nout=2, init=Xavier(), bias=Constant(0), activation=Logistic(shortcut=True))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyBinary())

    return (model, cost)