import numpy as np


class ConfusionMatrix:
    check_args = True
    classes = 0
    confusion = None
    needs_recalc = False

    def __init__(self, classes, check_args=True):
        self.classes = classes[:]
        self.check_args = check_args

    def check_segmentation(self, image_array):
        has_invalid_values = (np.any(image_array < 0) == True) or \
                             (np.any(image_array >= len(self.classes)) == True)
        if (has_invalid_values == True):
            raise Exception('Array contains unexpected values beyond interval [0; %d]: %s.' \
                % (len(self.classes) - 1, np.unique(image_array)))

    def update(self, outputs, targets):
        """
        Processes the given set of images and returns an IoU score.
        Adopted code from VOC 2012 mIoU calculation script.
        """

        if (self.confusion is None):
            self.confusion = np.zeros((len(self.classes), len(self.classes)))
        confusion = self.confusion

        output = outputs.reshape((-1))
        target = targets.reshape((-1))

        if (output.size != target.size):
            raise Exception("Array sizes are different: %s, %s" % (output.size, target.size))

        if (self.check_args == True):
            self.check_segmentation(output)
            self.check_segmentation(target)

        sumim = target + output * len(self.classes)
        hs, bin_edges = np.histogram(sumim[target < 255],
            bins=range(0, len(self.classes) * len(self.classes) + 1))
        confusion += hs.reshape(confusion.shape)

        self.needs_recalc = True

    def compute_overall_accuracy(self):
        self.overall_accuracy = 100 * \
            np.sum(np.diagonal(self.confusion)) / np.sum(self.confusion)
        return self.overall_accuracy

    def compute_class_accuracies(self):
        self.class_acc = np.zeros(len(self.classes))
        denoms = np.sum(self.confusion, axis=0)
        for i in range(len(self.classes)):
            if (denoms[i] == 0):
                denoms[i] = 1
            self.class_acc[i] = 100 * self.confusion[i][i] / denoms[i]
        self.aver_class_acc = np.sum(self.class_acc) / len(self.classes)
        return self.class_acc, self.aver_class_acc

    def compute_class_IoU(self):
        self.accuracies = np.zeros(len(self.classes))
        gtj = np.sum(self.confusion, axis=1)
        resj = np.sum(self.confusion, axis=0)
        for j in range(len(self.classes)):
            gtjresj = self.confusion[j][j]
            denom = gtj[j] + resj[j] - gtjresj
            if (denom == 0):
                denom = 1
            self.accuracies[j] = 100 * gtjresj / denom
        self.aver_accuracy = np.sum(self.accuracies) / len(self.classes)
        return self.accuracies, self.aver_accuracy

    def show_results(self):
        if (self.needs_recalc == True):
            self.compute_overall_accuracy()
            self.compute_class_accuracies()
            self.compute_class_IoU()
            self.needs_recalc = False

        print('----------------------------------')
        print('Percentage of pixels correctly labelled overall (mean accuracy): %6.3f%%' %
              self.overall_accuracy)
        print('----------------------------------')
        print('Mean class accuracy: %6.3f%%' % self.aver_class_acc)
        print('Class accuracies:')
        for i in range(len(self.classes)):
            print('  %18s: %6.3f%%' % (self.classes[i], self.class_acc[i]))
        print('----------------------------------')
        print('Mean IoU accuracy: %6.3f%%' % self.aver_accuracy)
        print('IoU class accuracies:')
        for i in range(len(self.classes)):
            print('  %18s: %6.3f%%' % (self.classes[i], self.accuracies[i]))
        print('----------------------------------')