from neon.transforms.cost import Metric
from models.confusion_matrix import ConfusionMatrix
from neon.layers.layer import Layer, ParameterLayer, BranchNode
from neon.layers.container import LayerContainer, flatten
from neon.layers.recurrent import interpret_in_shape, get_steps
import inspect
import numpy as np


def inverse_permutation(permutation):
    result = [0] * len(permutation)
    for i, v in enumerate(permutation):
        result[v] = i
    return result

def transpose_inplace(array, shuffle_dim):
    transposed = array.get().transpose(shuffle_dim)
    array = array.dimension_reorder(shuffle_dim)
    array[:] = np.ascontiguousarray(transposed)
    return array


class PixelwiseAccuracy(Metric):
    """
    Pixelwise mean accuracy metric (correct rate).

    """

    def __init__(self, c, w, h):
        super(PixelwiseAccuracy, self).__init__()

        self.preds = self.be.iobuf(h * w)
        self.hyps = self.be.iobuf(h * w)
        self.outputs = self.preds  # Contains per record metric
        self.c = c
        self.w = w
        self.h = h
        self.metric_names = ['Accuracy']

    def __call__(self, y_in, t_in, calcrange=slice(0, None)):
        """
        Returns the accuracy.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            float: Returns the metric
        """

        # (CHW, N) -> (C, HW, N) argmax (1, HW, N) -> (HW, N) -> (1, N) -> (1)
        y = y_in.reshape((self.c, self.h * self.w, -1))
        t = t_in.reshape((self.c, self.h * self.w, -1))
        # convert back from onehot and compare
        self.preds[:] = self.be.argmax(y, axis=0)
        self.hyps[:] = self.be.argmax(t, axis=0)
        self.outputs[:] = self.be.equal(self.preds, self.hyps)

        return self.outputs.get()[:, calcrange].mean()


class ConfusionMetric(Metric):
    confusion = None
    prediction = None
    target = None
    c = 0

    def __init__(self, classes):
        super(ConfusionMetric, self).__init__()

        self.confusion = ConfusionMatrix(classes)
        self.c = len(classes)
        self.metric_names = ['IoU', 'Accuracy']

    def __call__(self, outputs, targets, calcrange=slice(0, None)):
        """
        Returns the accuracy.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            0: Returns the dummy value
        """

        inner_dim = outputs.size // (self.c * self.be.bsz)  # h * w

        if (self.prediction is None):
            self.prediction = np.empty([inner_dim, self.be.bsz], dtype=float)
        if (self.target is None):
            self.target = np.empty([inner_dim, self.be.bsz], dtype=float)

        # (CHW, N) -> (C, HW, N) argmax(c)-> (1, HW, N) squeeze-> (HW, N)
        y = outputs.reshape((self.c, inner_dim, -1)).get()
        t = targets.reshape((self.c, inner_dim, -1)).get()
        # convert back from onehot and compare
        self.prediction[:] = np.argmax(y, axis=0)
        self.target[:] = np.argmax(t, axis=0)

        self.confusion.update(
            self.prediction[:, calcrange],
            self.target[:, calcrange])

        return 0

    def show_results(self):
        return self.confusion.show_results()

    def get_confusion(self):
        return self.confusion


class MSE(Metric):
    def __init__(self):
        super(Metric, self).__init__()
        self.buffer = None
        self.metric_names = ['MSE']

    def __call__(self, y, t, calcrange=slice(0, None)):
        # 1/N * sum( (y - t)^2 )
        if (self.buffer == None):
            self.buffer = self.be.iobuf(y.size / self.be.bsz)
        self.buffer[:] = self.be.square(self.be.subtract(y, t))
        return self.buffer.get()[:, calcrange].mean()


class ShapeTransform(Layer):
    def __init__(self, name=None):
        super(ShapeTransform, self).__init__(name)

        self.owns_output = False

    def __str__(self):
        s = "%s Layer '%s'" % (self.__class__.__name__, self.name)
        if ((self.in_shape is not None) and (self.out_shape is not None)):
            s += ": %s [%s] -> %s [%s]" % \
                (self.in_shape, self.in_shape_t,
                self.out_shape, self.out_shape_t)
        return s

    def _get_total_in_shape(self, in_shape):
        # Remember batch size
        in_shape_t = None
        if isinstance(in_shape, tuple):
            if len(in_shape) == 2:
                in_shape_t = (in_shape[0], in_shape[1] * self.be.bsz)
            else:
                in_shape_t = tuple([int(d) for d in in_shape] + [self.be.bsz])
        else:
            in_shape_t = (in_shape, self.be.bsz)
        return in_shape_t

    def _get_out_shape(self, out_shape_t):
        # Forget batch size
        out_shape = None
        if len(out_shape_t) == 2:
            out_shape = (out_shape_t[0], out_shape_t[1] // self.be.bsz)
        else:
            out_shape = tuple([int(d) for d in out_shape_t[:-1]])
        return out_shape

    def _configure_shape(self, in_shape_t):
        raise NotImplementedError

    def configure(self, in_obj):
        super(ShapeTransform, self).configure(in_obj)

        self.in_shape_t = self._get_total_in_shape(self.in_shape)
        self._configure_shape(self.in_shape_t)
        self.out_shape = self._get_out_shape(self.out_shape_t)

        return self

    def _fprop(self, inputs):
        raise NotImplementedError

    def fprop(self, inputs, inference=False):
        self.inputs = inputs
        self.outputs = self._fprop(self.inputs)
        return self.outputs

    def _bprop(self, error):
        raise NotImplementedError

    def bprop(self, error, alpha=1.0, beta=0.0):
        self._bprop(error)
        return self.deltas


class Extract(ShapeTransform):
    """
    Extracts elements at given indices.
    """

    def __init__(self, dim=None, indices=None, name=None):
        super(Extract, self).__init__(name)

        self.dim = dim
        self.indices = indices
        self.owns_output = True
        self.owns_delta = True

    def _configure_shape(self, in_shape_t):
        self.out_shape_t = tuple([len(self.indices)] + list(in_shape_t[self.dim + 1:]))

    def _fprop(self, x):
        self.outputs[:] = x[self.indices[0]] # .take(self.indices, self.dim) - neon cannot do this for some reason (not supported in 2.6.0)
        return self.outputs

    def _bprop(self, error):
        deltas_view = self.deltas.reshape(self.in_shape_t)
        error_view = error.reshape(self.out_shape_t)
        deltas_view[self.indices[0]][:] = error_view
        return self.deltas


class Reshape(ShapeTransform):
    """
    Reshapes inputs.
    """

    def __init__(self, fshape=None, name=None):
        super(Reshape, self).__init__(name)

        self.fshape = fshape # needed only for suppressing log messages 'neon - WARNING - can't describe argument 'fshape' to Reshape'
        self.out_shape_t = fshape

    def _configure_shape(self, in_shape_t):
        if -1 in self.out_shape_t:
            missing_dim = -int(np.prod(in_shape_t)) // int(np.prod(self.out_shape_t))
            self.out_shape_t = [missing_dim if x == -1 else x for x in self.out_shape_t]

    def _fprop(self, x):
        return x.reshape(self.out_shape_t)

    def _bprop(self, error):
        self.deltas = error.reshape(self.in_shape_t)
        return self.deltas


class DimShuffle(ShapeTransform):
    """
    Reorders input dimensions.
    """

    def __init__(self, shuffle_dim=None, name=None):
        super(DimShuffle, self).__init__(name)

        self.shuffle_dim = shuffle_dim
        self.inverse_shuffle = inverse_permutation(shuffle_dim)

    def _configure_shape(self, in_shape_t):
        self.out_shape_t = tuple([in_shape_t[d] for d in self.shuffle_dim])

    def _fprop(self, x):
        transpose_inplace(x, self.shuffle_dim)
        return x

    def _bprop(self, error):
        error = error.reshape(self.out_shape_t)
        transpose_inplace(error, self.inverse_shuffle)
        self.deltas = error
        return self.deltas


class Split(ShapeTransform):
    """
    Splits inputs by blocks.
    """

    def __init__(self, block_shape=None,
                 shuffle_dim=(0, 2, 4, 1, 3, 5, 6), name=None):
        super(Split, self).__init__(name)

        if ((block_shape is None) or (len(block_shape) != 3)):
            raise Exception('Expected a tuple-like of 3 ints for block_shape.')
        if ((shuffle_dim is None) or (len(shuffle_dim) != 7)):
            raise Exception(
                'Expected a tuple-like of 7 ints for shuffle order.')

        self.block_shape = block_shape
        self.shuffle_dim = shuffle_dim
        self.inverse_shuffle = inverse_permutation(shuffle_dim)

    def make_internal_shape(self, in_shape_t):
        internal_shape = in_shape_t[:-1]
        tiled_internal_shape = [0] * 2 * len(internal_shape) + [in_shape_t[-1]]
        for d in range(len(internal_shape)):
            tiled_internal_shape[2 * d + 0] = internal_shape[d] // self.block_shape[d]
            tiled_internal_shape[2 * d + 1] = self.block_shape[d]
        # (Co, Cb, Ho, Hb, Wo, Wb, N)
        return tuple(tiled_internal_shape)

    def _configure_shape(self, in_shape_t):
        self.internal_shape_t = self.make_internal_shape(in_shape_t)
        self.out_shape_t = tuple([self.internal_shape_t[d] for d in self.shuffle_dim])

    def _fprop(self, inputs):
        inputs = inputs.reshape(self.internal_shape_t)
        transpose_inplace(inputs, self.shuffle_dim)
        return inputs

    def _bprop(self, error):
        transpose_inplace(error, self.inverse_shuffle)
        self.deltas = error.reshape(self.in_shape_t)
        return self.deltas


def SpatialRNN(input_shape, block_shape, RNN, RNN_params):
    """
    In this block Bidirectional RNN is applied to
    each input image block (of shape block_shape).
    The results are concatenated.
    Output shape is:
      C = RNN outputs count,
      H, W = input_shape[1:] / block_shape[1:] (element-wise)
    """
    front_shape = [input_shape[d] // block_shape[d] for d in range(len(input_shape))]

    fC = front_shape[0]
    fH = front_shape[1]
    fW = front_shape[2]

    bC = block_shape[0]
    bH = block_shape[1]
    bW = block_shape[2]

    RNN_size = RNN_params['output_size']

    block_layers = [
        Split(block_shape=(bC, bH, bW), shuffle_dim=(4, 0, 1, 3, 5, 2, 6)),
        Reshape((fW, (fC * bC) * bH * bW, fH, -1))
    ]
    layers_vertical = []
    for i in range(fW):
        layers_vertical.append([
            Extract(dim=0, indices=[i]),
            RNN(**RNN_params),
            Reshape((2 * RNN_size, -1)) # workaround for MergeBroadcast incompatibility with RNN shape
        ])  # vertical image traversal
    block_layers.append(MergeBroadcast(layers_vertical, 'stack'))

    block_layers.extend([
        Reshape((fW, 2 * RNN_size, fH, -1)),  # 1st block output shape
        DimShuffle((2, 1, 0, 3)),
        Reshape((fH, 2 * RNN_size, fW, -1)),
    ])
    layers_horizontal = []
    for i in range(fH):
        layers_horizontal.append([
            Extract(dim=0, indices=[i]),
            RNN(**RNN_params),
            Reshape((2 * RNN_size, -1)) # workaround for MergeBroadcast incompatibility with RNN shape
        ])  # horizontal image traversal
    block_layers.append(MergeBroadcast(layers_horizontal, 'stack'))

    block_layers.extend([
        Reshape((fH, 2 * RNN_size, fW, -1)),  # 2nd block output shape
        DimShuffle((1, 0, 2, 3))
    ])

    return block_layers


# Layers with fixes

# Fixed allocation
class Sequential(LayerContainer):
    """
    Layer container that encapsulates a simple linear pathway of layers.

    Arguments:
        layers (list): List of objects which can be either a list of layers
                       (including layer containers).
    """
    def __init__(self, layers, name=None):
        super(Sequential, self).__init__(name)

        assert layers, "Provide layers"
        self.layers = [l for l in flatten(layers)]
        self._layers = [x for x in self.layers if type(x) not in (BranchNode,)]
        root = self._layers[0]
        assert (root.owns_output or
                type(root) in [Dropout, DataTransform]), "Sequential root must own outputs"

    def configure(self, in_obj):
        """
        Must receive a list of shapes for configuration (one for each pathway)
        the shapes correspond to the layer_container attribute

        Arguments:
            in_obj: any object that has an out_shape (Layer) or shape (Tensor, dataset)
        """
        if in_obj:
            config_layers = self.layers
            in_obj = in_obj
        else:
            in_obj = self.layers[0]

            # Remove the initial branch nodes from the layers
            for l_idx, l in enumerate(self.layers):
                if type(l) in (BranchNode,):
                    continue
                else:
                    config_layers = self.layers[l_idx:]
                    break

        super(Sequential, self).configure(in_obj)
        prev_layer = None
        for l in config_layers:
            in_obj = l.configure(in_obj)
            if prev_layer is not None:
                prev_layer.set_next(l)
            prev_layer = l
        self.parallelism = in_obj.parallelism
        self.out_shape = in_obj.out_shape
        return self

    def allocate(self, shared_outputs=None, accumulate_updates=False):
        """
        Allocate output buffer to store activations from fprop.

        Arguments:
            shared_outputs (Tensor, optional): pre-allocated tensor for activations to be
                                               computed into
        """
        # get the layers that own their outputs
        self.accumulate_updates = accumulate_updates
        alloc_layers = [l for l in self.layers if l.owns_output]
        if 'accumulate_updates' in inspect.getargspec(alloc_layers[-1].allocate).args:
            alloc_layers[-1].allocate(shared_outputs, accumulate_updates=accumulate_updates)
        else:
            alloc_layers[-1].allocate(shared_outputs)
        for l in [l for l in self.layers if l.owns_output][:-1]:
            if 'accumulate_updates' in inspect.getargspec(l.allocate).args:
                l.allocate(accumulate_updates=accumulate_updates)
            else:
                l.allocate()

    def allocate_deltas(self, global_deltas=None):
        if global_deltas is None:
            self.global_deltas = DeltasTree()

            st_ind = 0 if getattr(self.layers[0], 'nest_deltas', False) else 1
            for layer in self.layers[st_ind:]:
                layer.allocate_deltas(self.global_deltas)

            self.global_deltas.allocate_buffers()
        else:
            self.global_deltas = global_deltas

        self.set_deltas(self.global_deltas)

    def fprop(self, inputs, inference=False, beta=0.0):
        """
        TODO:  Handle final layers that don't own their own outputs (bias, activation)

        Arguments:
            inputs:
            inference:  (Default value = False)
            beta:  (Default value = 0.0)

        Returns:

        """
        x = inputs

        for l in self.layers:
            altered_tensor = l.be.distribute_data(x, l.parallelism)
            l.revert_list = [altered_tensor] if altered_tensor else []

            # try to convert to mkl
            l.be.convert_data(x, l.get_is_mklop())

            if l is self.layers[-1] and beta != 0:
                x = l.fprop(x, inference=inference, beta=beta)
            else:
                x = l.fprop(x, inference=inference)

        if inference:
            self.revert_tensors()

        return x

    def bprop(self, error, alpha=1.0, beta=0.0):
        """
        Apply the backward pass transformation to the input data.

        Arguments:
            error (Tensor): deltas back propagated from the adjacent higher layer
            alpha (float, optional): scale to apply to input for activation
                                     gradient bprop.  Defaults to 1.0
            beta (float, optional): scale to apply to output activation
                                    gradient bprop.  Defaults to 0.0

        Returns:
            Tensor: deltas to propagate to the adjacent lower layer
        """
        for l in reversed(self._layers):
            altered_tensor = l.be.distribute_data(error, l.parallelism)

            # try to convert to mkl
            l.be.convert_data(error, l.get_is_mklop())

            if altered_tensor:
                l.revert_list.append(altered_tensor)
            if type(l.prev_layer) is BranchNode or l is self._layers[0]:
                error = l.bprop(error, alpha, beta)
            else:
                error = l.bprop(error)

            # for not-mkl op, deltas is cpu tensor, but it is shared thus may have
            # meanless mkl tensor info, thus clean it for further operation (beta add)
            l.be.clean_data(l.deltas, not l.get_is_mklop())

            for tensor in l.revert_list:
                self.be.revert_tensor(tensor)
        return self._layers[0].deltas

    def get_terminal(self):
        """
        Used for recursively getting final nodes from layer containers.
        """
        terminal = self.layers[-1].get_terminal()
        return terminal


# Uses new Sequential
class Broadcast(LayerContainer):
    """
    Parent class for MergeSum and MergeBroadcast.
    """
    def __init__(self, layers, name=None):
        super(Broadcast, self).__init__(name)
        # Input list of layers converts:
        #   lists to Sequential container
        #   singleton layers to Sequential containers of 1
        #   leaves Sequentials alone
        self.layers = []
        for l in layers:
            if isinstance(l, Sequential):
                self.layers.append(l)
            elif isinstance(l, list):
                self.layers.append(Sequential(l))
            elif isinstance(l, Layer):
                self.layers.append(Sequential([l]))
            else:
                ValueError("Incompatible element for " + self.__class__.__name__ + " Layer")
        self.owns_output = True
        self.outputs = None

    @property
    def nest_deltas(self):
        return True

    def __str__(self):
        ss = '\n\t'.join([str(l) for l in self.layers])
        ss = '\t' + self.classnm + '\n\t' + ss
        return ss

    def configure(self, in_obj):
        """
        Sets shape based parameters of this layer given an input tuple or int
        or input layer

        Arguments:
            in_obj (int, tuple, Layer or Tensor or dataset): object that provides shape
                                                             information for layer

        Returns:
            (tuple): shape of output data
        """
        super(Broadcast, self).configure(in_obj)

        # Receiving from single source -- distribute to branches
        for l in self.layers:
            l.configure(in_obj)
        self._configure_merge()
        return self

    def allocate_deltas(self, global_deltas):
        nested_deltas = global_deltas.decend()
        for layer in self.layers:
            layer.layers[0].allocate_deltas(global_deltas)
            for sublayer in layer.layers[1:]:
                sublayer.allocate_deltas(nested_deltas)

    def set_deltas(self, delta_buffers):
        """
        Use pre-allocated (by layer containers) list of buffers for backpropagated error.
        Only set deltas for layers that own their own deltas
        Only allocate space if layer owns its own deltas (e.g., bias and activation work in-place,
        so do not own their deltas).

        Arguments:
            delta_buffers (DeltasTree): list of pre-allocated tensors (provided by layer container)
        """
        bottom_buffer = delta_buffers.buffers[0]

        nested_deltas = delta_buffers.decend()
        assert nested_deltas is not None
        for l in self.layers:
            l.layers[0].set_deltas(delta_buffers)

            # mkl need allocate new deltas
            l.layers[0].deltas = self.be.allocate_new_deltas(
                l.layers[0].deltas, l.layers[0].in_shape, l.layers[0].parallelism)

            delta_buffers.buffers.reverse()  # undo that last reverse
            for sublayer in l.layers[1:]:
                sublayer.set_deltas(nested_deltas)

        # Special case if originating from a branch node
        if type(self.prev_layer) is BranchNode:
            self.deltas = self.be.iobuf(self.in_shape, shared=self.prev_layer.deltas,
                                        parallelism=self.parallelism)
        else:
            self.deltas = self.be.iobuf(self.in_shape, shared=bottom_buffer,
                                        parallelism=self.parallelism)
            delta_buffers.buffers.reverse()

    def get_terminal(self):
        """
        Used for recursively getting final nodes from layer containers.
        """
        terminals = [l.get_terminal() for l in self.layers]
        return terminals


# Uses new Sequential
class MergeBroadcast(Broadcast):
    """
    Branches a single incoming layer or object (broadcast) into multiple output paths that are
    then combined again (merged). This container supports several options for concatenating the
    paths ("recurrent", "depth", and "stack").

    "recurrent" is used when merging two recurrent output streams.

    "depth" concatenates activations that have a notion of spatial dimension. Multiple
    activations can be concatenated along the feature map dimension, but the feature map
    shapes have to be the same.

    "stack" ignores the feature map shape and simply stacks the non-batch dimensions
    atop each other. Used to concatenate the output of fully connected layers with each
    other, and fully connected layers with convolutional layers.

    For example, suppose we are merging a conv layer with output shape (10, 5, 5)
    and a fully connected layer with 100 output nodes. Using 'depth' is not allowable.
    By using 'stack', the (10, 5, 5) output of the conv layer would just be interpreted as
    250 output nodes that are stacked on top of the 100 nodes from the fully connected
    layer to get a total merged output of 350 nodes.

    Arguments:
        layers (list(list(Layer), LayerContainer): list of either layer lists,
                                                   or layer containers.  Elements that are
                                                   lists will be wrapped in Sequential
                                                   containers
        merge (string): the merging method. Must be 'recurrent', 'depth', or 'stack'
        alphas (list(float), optional):  list of alpha values by which to weight the
                                         backpropagated errors
        name (str): Container name.  Defaults to "MergeBroadcast"
    """
    def __init__(self, layers, merge, alphas=None, name=None):
        super(MergeBroadcast, self).__init__(layers, name)

        self.betas = [1.0 for _ in self.layers]
        self.betas[-1] = 0.0
        self.alphas = [1.0 for _ in self.layers] if alphas is None else alphas

        self.merge = merge  # How this MergeBroadcast gets merged
        assert self.merge in ("recurrent", "depth", "stack")
        self.error_views = None
        self.ngLayer = self.be.mergebroadcast_layer(len(layers))

    def get_partitions(self, x, slices):
        """
        Given a partitioning, slices, of an activation buffer, x, determine which axis to slice
        along depending on whether x is a sequential tensor or not.

        Arguments:
            x:
            slices:

        Returns:

        """
        if x.shape[-1] != self.be.bsz:  # This is the sequential case
            return [x[:, sl] for sl in slices]
        else:
            return [x[sl] for sl in slices]

    def allocate(self, shared_outputs=None):
        """
        Allocate output buffer to store activations from fprop.

        Arguments:
            shared_outputs (Tensor, optional): pre-allocated tensor for activations to be
                                               computed into
        """
        if self.outputs is None:
            self.outputs = self.be.iobuf(self.out_shape, shared=shared_outputs,
                                         parallelism=self.parallelism)
        self.output_views = self.get_partitions(self.outputs, self.slices)
        for l, out_view in zip(self.layers, self.output_views):
            l.allocate(shared_outputs=out_view)

    def _configure_merge(self):
        """
        Helper function for configuring shapes depending on the merge concatenation type
        """
        in_shapes = [l.out_shape for l in self.layers]
        # Figure out how to merge
        if self.merge == "recurrent":
            catdims = [xs[1] for xs in in_shapes]
            self.out_shape = (in_shapes[0][0], sum(catdims))
            stride_size = self.be.bsz
        elif self.merge == "depth":
            catdims = [xs[0] for xs in in_shapes]
            self.out_shape = (sum(catdims),) + in_shapes[0][1:]
            stride_size = np.prod(in_shapes[0][1:])
        elif self.merge == "stack":
            catdims = [xs if isinstance(xs, int) else np.prod(xs) for xs in in_shapes]
            self.out_shape = sum(catdims)
            stride_size = 1
        end_idx = [idx * stride_size for idx in np.cumsum(catdims)]
        start_idx = [0] + end_idx[:-1]
        self.slices = [slice(s, e) for s, e in zip(start_idx, end_idx)]

    def fprop(self, inputs, inference=False):
        """
        Apply the forward pass transformation to the input data.

        Arguments:
            inputs (Tensor): input data

        Returns:
            Tensor: output data
        """
        self.be.fprop_mergebroadcast(
            self.ngLayer, inputs, inference, self.outputs,
            self.layers, self.out_shape)
        return self.outputs

    def bprop(self, error, alpha=1.0, beta=0.0):
        """
        Apply the backward pass transformation to the input data.

        Arguments:
            error (Tensor): deltas back propagated from the adjacent higher layer
            alpha (float, optional): scale to apply to input for activation
                                     gradient bprop.  Defaults to 1.0
            beta (float, optional): scale to apply to output activation
                                    gradient bprop.  Defaults to 0.0

        Returns:
            Tensor: deltas to propagate to the adjacent lower layer
        """
        if self.error_views is None:
            self.error_views = self.get_partitions(error, self.slices)
        self.be.bprop_mergebroadcast(
            self.ngLayer, self.layers, self.error_views, error,
            self.deltas, self.out_shape, alpha, beta, self.alphas, self.betas)
        return self.deltas


# Fixed allocation
class BiRNN(ParameterLayer):

    """
    Basic Bi Directional Recurrent layer.

    Arguments:
        output_size (int): Number of hidden/output units
        init (Initializer): Function for initializing the model parameters
        init_inner (Initializer, optional): Function for initializing the model's recurrent
                                            parameters.  If absent, will default to using same
                                            initializer provided to init.
        activation (Transform): Activation function for the input modulation
        reset_cells (bool): default to be False to make the layer stateful,
                            set to True to be stateless.
        split_inputs (bool): to expect the input coming from the same source of separate
                             sources
        name (str, optional): name to refer to this layer as.

    Attributes:
        W_input (Tensor): weights from inputs to output units
            (input_size, output_size)
        W_recur (Tensor): weights for recurrent connections
            (output_size, output_size)
        b (Tensor): Biases on output units (output_size, 1)
    """

    def __init__(self, output_size, init, init_inner=None, activation=None,
                 reset_cells=False, split_inputs=False, name=None,
                 parallelism="Data"):
        super(BiRNN, self).__init__(init, name, parallelism=parallelism)
        self.in_deltas_f = None
        self.in_deltas_b = None
        self.nout = output_size
        self.h_nout = output_size
        self.output_size = output_size
        self.activation = activation
        self.h_buffer = None
        self.W_input = None
        self.ngates = 1
        self.split_inputs = split_inputs
        self.reset_cells = reset_cells
        self.init_inner = init_inner
        self.x = None

    def __str__(self):
        if self.split_inputs:
            return "BiRNN Layer '%s': (%d inputs) * 2, (%d outputs) * 2, %d steps" % (
                self.name, self.nin // 2, self.nout, self.nsteps)
        else:
            return "BiRNN Layer '%s': %d inputs, (%d outputs) * 2, %d steps" % (
                self.name, self.nin, self.nout, self.nsteps)

    def configure(self, in_obj):
        """
        Set shape based parameters of this layer given an input tuple, int
        or input layer.

        Arguments:
            in_obj (int, tuple, Layer, Tensor or dataset): object that provides shape
                                                           information for layer

        Returns:
            (tuple): shape of output data
        """
        super(BiRNN, self).configure(in_obj)
        (self.nin, self.nsteps) = interpret_in_shape(self.in_shape)

        self.out_shape = (2 * self.nout, self.nsteps)
        self.gate_shape = (2 * self.nout * self.ngates, self.nsteps)
        self.hidden_shape = (2 * self.nout, self.nsteps + 2)

        if self.split_inputs is True and self.nin % 2 == 1:
            raise ValueError("# inputs units is odd and split_inputs is True ")

        self.o_shape = (self.nout, self.nsteps)
        self.g_shape = (self.nout * self.ngates, self.nsteps)
        self.i_shape = (
            self.nin // 2, self.nsteps) if self.split_inputs else (self.nin, self.nsteps)

        if self.weight_shape is None:
            self.weight_shape = (self.nout, self.nin)

        return self

    def allocate(self, shared_outputs=None):
        """
        Allocate output buffer to store activations from fprop.

        Arguments:
            shared_outputs (Tensor, optional): pre-allocated tensor for activations to be
                                               computed into
        """
        assert self.owns_output

        # Allocate a hidden state buffer with an extra time step of padding on either end
        o_shape_pad = (self.o_shape[0], self.o_shape[1] + 2)
        self.h_buffer_all = self.be.iobuf(self.hidden_shape,
                                          shared=None,
                                          parallelism=self.parallelism)

        # bugfix in outputs allocation
        self.output_buffer = self.be.iobuf(self.out_shape,
                                          shared=shared_outputs,
                                          parallelism=self.parallelism)

        step_size = self.h_buffer_all.shape[1] // (self.nsteps + 2)
        self.outputs = self.h_buffer_all[:, step_size:-step_size]
        super(BiRNN, self).allocate(shared_outputs)

        nout = self.o_shape[0]
        self.h_buffer = self.outputs
        self.out_deltas_buffer = self.deltas

        # forward
        self.h_buffer_f = self.h_buffer[:nout]
        self.h_prev_buffer = self.h_buffer_all[:nout, :-(2 * step_size)]
        self.h_f_last = self.h_prev_buffer[:, :step_size]

        # Slice padded buffer to a list of step tensors, remove padding steps
        self.h_f = get_steps(self.h_buffer_all[:nout, :], o_shape_pad)[1:-1]

        # Slice padded buffer to a list of step tensors, remove the last step and end padding
        self.h_prev = get_steps(self.h_buffer_all[:nout, :], o_shape_pad)[:-2]

        # backward
        self.h_buffer_b = self.h_buffer[nout:]
        self.h_next_buffer = self.h_buffer_all[nout:, (2 * step_size):]
        self.h_b_last = self.h_next_buffer[:, -step_size:]

        # Slice padded buffer to a list of step tensors, remove padding steps
        self.h_b = get_steps(self.h_buffer_all[nout:, :], o_shape_pad)[1:-1]

        # Slice padded buffer to a list of step tensors, remove the beginning padding and 1st step
        self.h_next = get_steps(self.h_buffer_all[nout:, :], o_shape_pad)[2:]

        self.bufs_to_reset = [self.h_buffer]

        # Last time step for in deltas
        self.prev_in_deltas_last = self.be.empty_like(self.h_f[-1])
        self.next_in_deltas_last = self.be.empty_like(self.h_b[0])

        if self.W_input is None:
            self.init_params(self.weight_shape)

    def set_deltas(self, delta_buffers):
        """
        Use pre-allocated (by layer containers) list of buffers for backpropagated error.
        Only set deltas for layers that own their own deltas
        Only allocate space if layer owns its own deltas (e.g., bias and activation work in-place,
        so do not own their deltas).

        Arguments:
            delta_buffers (list): list of pre-allocated tensors (provided by layer container)
        """
        super(BiRNN, self).set_deltas(delta_buffers)
        self.out_deltas_buffer = self.deltas
        nin = self.i_shape[0]
        if self.split_inputs:
            self.out_deltas_buffer_f = self.out_deltas_buffer[:nin]
            self.out_deltas_buffer_b = self.out_deltas_buffer[nin:]
        else:
            self.out_deltas_buffer_f = self.out_deltas_buffer
            self.out_deltas_buffer_b = self.out_deltas_buffer
        self.out_delta_f = get_steps(self.out_deltas_buffer_f, self.i_shape)
        self.out_delta_b = get_steps(self.out_deltas_buffer_b, self.i_shape)
        self.out_deltas_buffer_f_v = self.out_deltas_buffer_f.reshape(nin, -1)
        self.out_deltas_buffer_b_v = self.out_deltas_buffer_b.reshape(nin, -1)

    def init_buffers(self, inputs):
        """
        Initialize buffers for recurrent internal units and outputs.
        Buffers are initialized as 2D tensors with second dimension being steps * batch_size
        A list of views are created on the buffer for easy manipulation of data
        related to a certain time step

        Arguments:
            inputs (Tensor): input data as 2D tensor. The dimension is
                             (input_size, sequence_length * batch_size)

        """
        if self.x is None or self.x.base is not inputs:
            if self.x:
                for buf in self.bufs_to_reset:
                    buf[:] = 0

            assert inputs.size == self.nin * self.nsteps * self.be.bsz

            self.x = inputs.reshape(self.nin, self.nsteps * self.be.bsz)
            nin = self.i_shape[0]
            if self.split_inputs:
                self.x_f = self.x[:nin]
                self.x_b = self.x[nin:]
            else:
                self.x_f = self.x
                self.x_b = self.x

            self.xs_f = get_steps(self.x_f, self.i_shape)
            self.xs_b = get_steps(self.x_b, self.i_shape)
            self.x_f_v = self.x_f.reshape(nin, -1)
            self.x_b_v = self.x_b.reshape(nin, -1)

    def init_params(self, shape):
        """
        Initialize params for LSTM including weights and biases.
        The weight matrix and bias matrix are concatenated from the weights
        for inputs and weights for recurrent inputs and bias.
        The shape of the weights are (number of inputs + number of outputs +1 )
        by (number of outputs * 4)

        Arguments:
            shape (Tuple): contains number of outputs and number of inputs

        """
        (nout, nin) = (self.o_shape[0], self.i_shape[0])
        self.g_nout = self.ngates * nout
        Wshape = (2 * (nin + nout + 1), self.g_nout)
        doFill = False

        # Weights: input, recurrent, bias
        if self.W is None:
            self.W = self.be.empty(Wshape, **self.get_param_attrs())
            self.dW = self.be.zeros_like(self.W)
            doFill = True
        else:
            # Deserialized weights and empty grad
            assert self.W.shape == Wshape
            assert self.dW.shape == Wshape

        self.W_input_f = self.W[:nin].reshape((self.g_nout, nin))
        self.W_input_b = self.W[nin:2 * nin].reshape((self.g_nout, nin))

        self.W_recur_f = self.W[2 * nin:2 * nin + nout].reshape((self.g_nout, nout))
        self.W_recur_b = self.W[
            2 * nin + nout:2 * nin + 2 * nout].reshape((self.g_nout, nout))

        self.b_f = self.W[-2:-1].reshape((self.g_nout, 1))
        self.b_b = self.W[-1:].reshape((self.g_nout, 1))

        self.dW_input_f = self.dW[:nin].reshape(self.W_input_f.shape)
        self.dW_input_b = self.dW[nin:2 * nin].reshape(self.W_input_b.shape)

        self.dW_recur_f = self.dW[
            2 * nin:2 * nin + nout].reshape(self.W_recur_f.shape)
        self.dW_recur_b = self.dW[
            2 * nin + nout:2 * nin + 2 * nout].reshape(self.W_recur_b.shape)

        self.db_f = self.dW[-2:-1].reshape(self.b_f.shape)
        self.db_b = self.dW[-1:].reshape(self.b_b.shape)

        if doFill:
            gatelist = [g * nout for g in range(0, self.ngates + 1)]
            for wtnm in ('W_input_f', 'W_input_b', 'W_recur_f', 'W_recur_b'):
                wtmat = getattr(self, wtnm)
                if 'W_recur' in wtnm and self.init_inner is not None:
                    initfunc = self.init_inner
                else:
                    initfunc = self.init

                for gb, ge in zip(gatelist[:-1], gatelist[1:]):
                    initfunc.fill(wtmat[gb:ge])
            self.b_f.fill(0.)
            self.b_b.fill(0.)

    def fprop(self, inputs, inference=False):
        """
        Forward propagation of input to bi-directional recurrent layer.

        Arguments:
            inputs (Tensor): input to the model for each time step of
                             unrolling for each input in minibatch
                             shape: (feature_size, sequence_length * batch_size)
                             where:

                             * feature_size: input size
                             * sequence_length: degree of model unrolling
                             * batch_size: number of inputs in each mini-batch

            inference (bool, optional): Set to true if you are running
                                        inference (only care about forward
                                        propagation without associated backward
                                        propagation).  Default is False.

        Returns:
            Tensor: layer output activations for each time step of
                unrolling and for each input in the minibatch
                shape: (output_size, sequence_length * batch_size)
        """

        self.init_buffers(inputs)

        if self.reset_cells:
            self.h_f_last[:] = 0
            self.h_b_last[:] = 0
        else:
            self.h_f_last[:] = self.h_f[-1]
            self.h_b_last[:] = self.h_b[0]

        # Use single multiply for W_input
        self.be.compound_dot(self.W_input_f, self.x_f_v, self.h_buffer_f)
        self.be.compound_dot(self.W_input_b, self.x_b_v, self.h_buffer_b)

        self.be.compound_rnn_unroll_fprop(self.W_recur_f, self.h_prev,
                                          self.h_f, self.h_f, self.b_f,
                                          self.nout, self.nsteps,
                                          self.nsteps,
                                          self.activation, False)

        self.be.compound_rnn_unroll_fprop(self.W_recur_b, self.h_next,
                                          self.h_b, self.h_b, self.b_b,
                                          self.nout, self.nsteps,
                                          self.nsteps,
                                          self.activation, True)

        self.output_buffer[:] = self.h_buffer
        return self.h_buffer

    def bprop(self, error, alpha=1.0, beta=1.0):
        """
        Backward propagation of errors through bi-directional recurrent layer.

        Arguments:
            deltas (Tensor): tensors containing the errors for
                each step of model unrolling.
                shape: (output_size, sequence_length * batch_size)
            alpha (float, optional): scale to apply to input for activation
                                     gradient bprop.  Defaults to 1.0
            beta (float, optional): scale to apply to output activation
                                    gradient bprop.  Defaults to 0.0

        Returns:
            Tensor: back propagated errors for each step of time unrolling
                for each mini-batch element
                shape: (input_size, sequence_length * batch_size)
        """
        if self.in_deltas_f is None:
            self.in_deltas_f = get_steps(error[:self.nout], self.o_shape)
            self.prev_in_deltas = [self.prev_in_deltas_last] + self.in_deltas_f[:-1]

        if self.in_deltas_b is None:
            self.in_deltas_b = get_steps(error[self.nout:], self.o_shape)
            self.next_in_deltas = self.in_deltas_b[1:] + [self.next_in_deltas_last]
        self.out_deltas_buffer[:] = 0

        self.be.compound_rnn_unroll_bprop(self.W_recur_f.T, self.prev_in_deltas,
                                          self.in_deltas_f, self.h_f,
                                          self.nout, self.nsteps,
                                          self.nsteps,
                                          self.activation, True)

        self.be.compound_rnn_unroll_bprop(self.W_recur_b.T, self.next_in_deltas,
                                          self.in_deltas_b, self.h_b,
                                          self.nout, self.nsteps,
                                          self.nsteps,
                                          self.activation, False)

        # Update gradients and output deltas for forward pass
        in_deltas_all_f = error[:self.nout]
        in_deltas_cur_f = in_deltas_all_f[:, self.be.bsz:]
        h_prev_all = self.h_buffer_f[:, :-self.be.bsz]
        self.be.compound_dot(in_deltas_cur_f, h_prev_all.T,
                             self.dW_recur_f)

        # Update gradients and output deltas for backward pass
        in_deltas_all_b = error[self.nout:]
        in_deltas_cur_b = in_deltas_all_b[:, :-self.be.bsz]
        h_next_all = self.h_buffer_b[:, self.be.bsz:]
        self.be.compound_dot(in_deltas_cur_b, h_next_all.T, self.dW_recur_b)

        self.be.compound_dot(in_deltas_all_f, self.x_f_v.T, self.dW_input_f)
        self.db_f[:] = self.be.sum(in_deltas_all_f, axis=1)
        if self.out_deltas_buffer_f:
            self.be.compound_dot(self.W_input_f.T, in_deltas_all_f,
                                 self.out_deltas_buffer_f_v,
                                 alpha=alpha, beta=beta)

        self.be.compound_dot(in_deltas_all_b, self.x_b_v.T, self.dW_input_b)
        self.db_b[:] = self.be.sum(in_deltas_all_b, axis=1)
        if self.out_deltas_buffer_b:
            self.be.compound_dot(self.W_input_b.T, in_deltas_all_b,
                                 self.out_deltas_buffer_b_v,
                                 alpha=alpha, beta=beta)

        return self.out_deltas_buffer
