from neon.initializers import Gaussian, Kaiming, Constant, Xavier, GlorotUniform
from neon.layers import GeneralizedCost, Affine, Conv, Deconv, Pooling, BatchNorm, Activation, Dropout, DataTransform, Linear, Bias, MergeSum, SkipNode
from neon.transforms import Rectlin, Logistic, Softmax, PixelwiseSoftmax, Tanh, CrossEntropyBinary, CrossEntropyMulti, SumSquared, Normalizer, Explin
from neon.models import Model
from models.layers import SpatialRNN, BiRNN


def generate_cnn1_seg_model():
    # lr 0.00001

    class_count = 32

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # convolutional encoder / feature extractor
        # resolution 1
        Conv(fshape=(3, 3, 32), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),

        # resolution 1/2
        Conv(fshape=(3, 3, 64), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),

        Deconv(fshape=(4, 4, 64), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 32), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),

        # classifier
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0),
            activation=PixelwiseSoftmax(class_count))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyMulti())

    return (model, cost)

def generate_rnn1_cls_model(input_shape=(3, 128, 128)):
    iC = input_shape[0]
    iH = input_shape[1]
    iW = input_shape[2]

    class_count = 2

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # convolutional encoder / feature extractor
        # resolution 1
        BatchNorm(),
        Conv(fshape=(3, 3, 32), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/2
        Conv(fshape=(3, 3, 64), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/4
        Conv(fshape=(3, 3, 128), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/8
        Conv(fshape=(3, 3, 256), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        SpatialRNN(input_shape=(256, iH // 8, iW // 8), block_shape=(256, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 256, 'init': GlorotUniform(), 'activation': Tanh()}
        ), # outputs: (2 * 256, iH // 16, iW // 16, N)
        # # resolution 1/16

        BatchNorm(),
        SpatialRNN(input_shape=(512, iH // 16, iW // 16), block_shape=(512, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 512, 'init': GlorotUniform(), 'activation': Tanh()}
        ), # outputs: (2 * 512, iH // 32, iW // 32, N)
        # # resolution 1/32

        BatchNorm(),
        Conv(fshape=(1, 1, class_count), padding=1, strides=1, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        Pooling(fshape='all', padding=0, strides=1, op='avg'),
        Activation(Softmax())
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyMulti())

    return (model, cost)

def generate_rnn1_seg_model(input_shape=(3, 256, 256)):
    iC = input_shape[0]
    iH = input_shape[1]
    iW = input_shape[2]

    class_count = 32

    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        # convolutional encoder / feature extractor
        # resolution 1
        BatchNorm(),
        Conv(fshape=(3, 3, 32), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/2
        Conv(fshape=(3, 3, 64), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/4
        Conv(fshape=(3, 3, 128), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max'),
        # resolution 1/8
        Conv(fshape=(3, 3, 256), padding=2, strides=1, dilation=2, init=Kaiming(), bias=Constant(0), activation=Rectlin()),

        BatchNorm(),
        SpatialRNN(input_shape=(256, iH // 8, iW // 8), block_shape=(256, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 256, 'init': GlorotUniform(), 'activation': Tanh()}
        ), # outputs: (2 * 256, iH // 16, iW // 16, N)
        # # resolution 1/16

        BatchNorm(),
        SpatialRNN(input_shape=(512, iH // 16, iW // 16), block_shape=(512, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 512, 'init': GlorotUniform(), 'activation': Tanh()}
        ), # outputs: (2 * 512, iH // 32, iW // 32, N)
        # # resolution 1/32

        BatchNorm(),
        Deconv(fshape=(4, 4, 512), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 256), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 128), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 64), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 32), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),

        # classifier
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0),
            activation=PixelwiseSoftmax(class_count))
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyMulti())

    return (model, cost)

def generate_vgg_16_seg_model():
    class_count = 32

    init1 = Xavier(local=True)
    initfc = GlorotUniform()

    relu = Rectlin()
    conv_params = {'init': init1,
                   'strides': 1,
                   'padding': 1}

    # Set up the model layers
    layers = []

    # set up 3x3 conv stacks with different feature map sizes
    for i, nofm in enumerate([64, 128, 256, 512, 512]):
        i = i + 1
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_1' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_1_bias' % i))
        layers.append(Activation(Rectlin()))
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_2' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_2_bias' % i))
        layers.append(Activation(Rectlin()))
        if nofm > 128:
            layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_3' % i))
            layers.append(Bias(init=Constant(0), name='conv%d_3_bias' % i))
            layers.append(Activation(Rectlin()))
        layers.append(Pooling(2, strides=2))

    layers.extend([
        # resolution 1/32

        BatchNorm(),
        Deconv(fshape=(4, 4, 512), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 256), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 128), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 64), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 32), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),

        # classifier
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0),
            activation=PixelwiseSoftmax(class_count))
    ])

    model = Model(layers=layers)
    cost = GeneralizedCost(costfunc=CrossEntropyMulti())
    return (model, cost)

def generate_vgg_16_rnn1_seg_model(input_shape=(3, 256, 256)):
    iC = input_shape[0]
    iH = input_shape[1]
    iW = input_shape[2]

    class_count = 32

    init1 = Xavier(local=True)
    initfc = GlorotUniform()

    relu = Rectlin()
    conv_params = {'init': init1,
                   'strides': 1,
                   'padding': 1}

    # Set up the model layers
    layers = []

    # set up 3x3 conv stacks with different feature map sizes
    for i, nofm in enumerate([64, 128, 256, 512, 512]):
        i = i + 1
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_1' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_1_bias' % i))
        layers.append(Activation(Rectlin()))
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_2' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_2_bias' % i))
        layers.append(Activation(Rectlin()))
        if nofm > 128:
            layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_3' % i))
            layers.append(Bias(init=Constant(0), name='conv%d_3_bias' % i))
            layers.append(Activation(Rectlin()))
        layers.append(Pooling(2, strides=2))

    layers.extend([
        # resolution 1/32

        BatchNorm(),
        SpatialRNN(input_shape=(512, iH // 32, iW // 32), block_shape=(512, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 512, 'init': GlorotUniform(), 'activation': Tanh()}
        ),
        # resolution 1/64

        BatchNorm(),
        SpatialRNN(input_shape=(1024, iH // 64, iW // 64), block_shape=(1024, 2, 2),
            RNN=BiRNN, RNN_params={'output_size': 1024, 'init': GlorotUniform(), 'activation': Tanh()}
        ),
        # resolution 1/128

        BatchNorm(),
        SpatialRNN(input_shape=(2048, iH // 128, iW // 128), block_shape=(2048, 1, 1),
            RNN=BiRNN, RNN_params={'output_size': 2048, 'init': GlorotUniform(), 'activation': Tanh()}
        ),
        # resolution 1/128

        BatchNorm(),
        Deconv(fshape=(4, 4, 2048), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 1024), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 512), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 256), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 128), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 64), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),
        Deconv(fshape=(4, 4, 32), padding=1, strides=2, dilation=1, init=Kaiming(), bias=Constant(0), activation=Rectlin()),
        BatchNorm(),

        # classifier
        Conv(fshape=(1, 1, class_count), padding=0, strides=1, dilation=1, init=Kaiming(), bias=Constant(0),
            activation=PixelwiseSoftmax(class_count))
    ])

    model = Model(layers=layers)
    cost = GeneralizedCost(costfunc=CrossEntropyMulti())
    return (model, cost)