from neon.initializers import Gaussian, Kaiming, Constant, Xavier, GlorotUniform
from neon.layers import GeneralizedCost, Affine, Conv, Deconv, Pooling, BatchNorm, Activation, Dropout, DataTransform, Linear, Bias, MergeSum, SkipNode
from neon.transforms import Rectlin, Logistic, Softmax, Tanh, CrossEntropyBinary, CrossEntropyMulti, Normalizer
from neon.models import Model


def generate_mlp2b2_transfer_model():
    # This model is intended to be used in transfer learning
    # as the successor of the simplier 'mlp' model.
    # Note that to make trasferring work we should introduce new
    # names for the layers we don't want to copy from the previous model.
    layers = [
        DataTransform(transform=Normalizer(divisor=128.0)),

        Affine(nout=16, init=Gaussian(scale=0.1), bias=Constant(0),
            activation=Tanh()),
        Affine(nout=32, init=Gaussian(scale=0.1), bias=Constant(0),
            activation=Tanh()),
        Affine(nout=32, init=Gaussian(scale=0.1), bias=Constant(0),
            activation=Tanh(), name="l3"),
        Affine(nout=2, init=Gaussian(scale=0.1), bias=Constant(0),
            activation=Logistic(shortcut=True), name="cls")
    ]

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyBinary())

    return (model, cost)


def generate_alexnet_transfer_model():
    # Source model might be obtained here:
    # https://github.com/NervanaSystems/ModelZoo/tree/master/ImageClassification/ILSVRC2012/Alexnet
    # Direct link: https://s3-us-west-1.amazonaws.com/nervana-modelzoo/alexnet/alexnet.p
    #
    # To adopt this model to IMDB-WIKI face classification task,
    # we make some changes in the model, leaving saved parts unchanged.
    # The newly added layers must have different names from source model layers.
    # Learning parameters: base lr: 0.001, 5 epochs, batch size 128, wdecay 0.0005
    from neon.layers import LRN

    init_g1 = Gaussian(scale=0.01)
    init_g2 = Gaussian(scale=0.005)
    relu = Rectlin()

    layers = [
        # Saved layers:
        Conv((11, 11, 96), padding=0, strides=4, init=init_g1, name='conv1'),
        Bias(init=Constant(0), name='conv1_bias'),
        Activation(Rectlin(), name='conv1_Rectlin'),
        Pooling(3, strides=2, name='pool1'),
        LRN(5, ascale=0.0001, bpower=0.75, name='norm1'),
        Conv((5, 5, 256), padding=2, init=init_g1, name='conv2'),
        Bias(init=Constant(0), name='conv2_bias'),
        Activation(Rectlin(), name='conv2_Rectlin'),
        Pooling(3, strides=2, name='pool2'),
        LRN(5, ascale=0.0001, bpower=0.75, name='norm2'),
        Conv((3, 3, 384), padding=1, init=init_g1, name='conv3'),
        Bias(init=Constant(0), name='conv3_bias'),
        Activation(Rectlin(), name='conv3_Rectlin'),
        Conv((3, 3, 384), padding=1, init=init_g1, name='conv4'),
        Bias(init=Constant(1.0), name='conv4_bias'),
        Activation(Rectlin(), name='conv4_Rectlin'),
        Conv((3, 3, 256), padding=1, init=init_g1, name='conv5'),
        Bias(init=Constant(1.0), name='conv5_bias'),
        Activation(Rectlin(), name='conv5_Rectlin'),
        Pooling(3, strides=2, name='pool5'),

        # New layers:
        Affine(nout=4096, init=init_g2, bias=Constant(1.0),
            activation=relu, name='fc6_new'), # new since dimensions not match
        Dropout(keep=0.5, name='drop6'),
        Affine(nout=4096, init=init_g2, bias=Constant(1.0),
            activation=relu, name='fc7'),
        Dropout(keep=0.5, name='drop7'),
        Affine(nout=2, init=Xavier(), bias=Constant(0), activation=Softmax(),
            name='cls_imdb_wiki_face'),
    ]

    return Model(layers=layers), GeneralizedCost(costfunc=CrossEntropyMulti())


def generate_vgg_16_transfer_model():
    # Code from
    # https://github.com/NervanaSystems/ModelZoo/tree/master/ImageClassification/ILSVRC2012/VGG
    # weights from: https://s3-us-west-1.amazonaws.com/nervana-modelzoo/VGG/VGG_D.p

    init1 = Xavier(local=True)
    initfc = GlorotUniform()

    relu = Rectlin()
    conv_params = {'init': init1,
                   'strides': 1,
                   'padding': 1}

    # Set up the model layers
    layers = []

    # set up 3x3 conv stacks with different feature map sizes
    for i, nofm in enumerate([64, 128, 256, 512, 512]):
        i = i + 1
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_1' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_1_bias' % i))
        layers.append(Activation(Rectlin()))
        layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_2' % i))
        layers.append(Bias(init=Constant(0), name='conv%d_2_bias' % i))
        layers.append(Activation(Rectlin()))
        if nofm > 128:
            layers.append(Conv((3, 3, nofm), **conv_params, name='conv%d_3' % i))
            layers.append(Bias(init=Constant(0), name='conv%d_3_bias' % i))
            layers.append(Activation(Rectlin()))
        layers.append(Pooling(2, strides=2))

    layers.append(Affine(nout=4096, init=initfc, bias=Constant(0), activation=Rectlin(), name='fc6_new'))
    layers.append(Dropout(keep=0.5))
    layers.append(Affine(nout=4096, init=initfc, bias=Constant(0), activation=Rectlin(), name='fc7'))
    layers.append(Dropout(keep=0.5))
    layers.append(Affine(nout=2, init=initfc, bias=Constant(0), activation=Softmax(),
        name='cls_imdb_wiki_face'))

    model = Model(layers=layers)
    cost = GeneralizedCost(costfunc=CrossEntropyMulti())
    return (model, cost)


def make_resnet_base(stage_depths, block='basic', base_channels=64):
    '''
        Returns ResNet layers without the classifier layers.
        Used pre-activations model variant.
    '''
    # Implementation from
    # https://github.com/NervanaSystems/neon/blob/master/examples/cifar10_msra/network.py
    # Augmented with basic blocks and different stage sizes support.
    from math import log2

    def conv_params(fsize, nfm, stride=1, relu=True, batch_norm=True):
        return dict(
            fshape=(fsize, fsize, nfm),
            strides=stride,
            padding=fsize // 2,
            activation=(Rectlin() if relu else None),
            init=Kaiming(local=True),
            batch_norm=batch_norm)

    def module_basic(nfm, first=False, stride=1):
        # building block for ResNet-18, -34

        mainpath = [] if first else [BatchNorm(), Activation(Rectlin())]
        mainpath.append(Conv(**conv_params(3, nfm, stride=stride)))
        mainpath.append(Conv(**conv_params(3, nfm, relu=False, batch_norm=False)))

        sidepath = Conv(**conv_params(1, nfm, stride=stride, relu=False, batch_norm=False)) if (first or (stride != 1)) else SkipNode()

        return MergeSum([sidepath, mainpath])

    def module_bottleneck(nfm, first=False, stride=1):
        # building block for ResNet-50, -101, -152

        mainpath = [] if first else [BatchNorm(), Activation(Rectlin())]
        mainpath.append(Conv(**conv_params(1, nfm, stride=stride)))
        mainpath.append(Conv(**conv_params(3, nfm)))
        mainpath.append(Conv(**conv_params(1, nfm * 4, relu=False, batch_norm=False)))

        sidepath = Conv(**conv_params(1, nfm * 4, stride=stride, relu=False, batch_norm=False)) if (first or (stride != 1)) else SkipNode()

        return MergeSum([sidepath, mainpath])

    blocks = {'basic': module_basic, 'bottleneck': module_bottleneck}

    # Structure of the deep residual part of the network:
    # stage_depth modules of 2 convolutional layers each at feature map depths of 16, 32, 64
    # nfms = [2**(stage + 4) for stage in sorted(list(range(3)) * stage_depth)]

    # allowing different stage depths
    # so all the ResNet configurations from the paper become available
    # 18:  [2, 2, 2, 2], output features = 512, block = 'basic'
    # 34:  [3, 4, 6, 3], output features = 512, block = 'basic'
    # 50:  [3, 4, 6, 3], output features = 2048, block = 'bottleneck'
    # 101: [3, 4, 23, 3], output features = 2048, block = 'bottleneck'
    # 152: [3, 8, 36, 3], output features = 2048, block = 'bottleneck'
    stage_depths_populated = []
    for stage, depth in enumerate(stage_depths):
        stage_depths_populated.extend([stage] * depth)
    # nfms is a list of channel counts in blocks
    nfms = [2**(stage + int(log2(base_channels))) for stage in stage_depths_populated]

    strides = [1 if cur == prev else 2 for cur, prev in zip(nfms[1:], nfms[:-1])]
    module = blocks[block]

    # Now construct the network
    layers = [
        Conv(**conv_params(7, base_channels, 2)),
        Pooling(fshape=(3, 3), padding=1, strides=2, op='max')
    ]
    layers.append(module(nfms[0], first=True))
    for nfm, stride in zip(nfms[1:], strides):
        layers.append(module(nfm, stride=stride))

    return layers

def generate_resnet18_model():
    layers = make_resnet_base([2, 2, 2, 2], block='basic')
    layers.append(BatchNorm())
    layers.append(Activation(Rectlin()))
    layers.append(Pooling('all', op='avg'))
    layers.append(Affine(2, init=Kaiming(local=False), activation=Softmax()))

    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyMulti())

    return (model, cost)

def generate_resnet50_model():
    layers = make_resnet_base([3, 4, 6, 3], block='bottleneck')
    layers.append(BatchNorm())
    layers.append(Activation(Rectlin()))
    layers.append(Pooling('all', op='avg'))
    layers.append(Affine(2, init=Kaiming(local=False), activation=Softmax()))
    model = Model(layers=layers)

    cost = GeneralizedCost(costfunc=CrossEntropyMulti())

    return (model, cost)


def is_layer_matches(source_layer, target_layer):
    try:
        if (source_layer is None):
            raise Exception("Source layer is undefined")
        if (target_layer is None):
            raise Exception("Target layer is undefined")

        source_name = source_layer['config']['name']
        target_name = target_layer['config']['name']
        if ((len(source_name) == 0) or (len(target_name) == 0)):
            raise Exception("Layer names mismatch")
        if (source_name != target_name):
            raise Exception("Layer names mismatch (%s -> %s)" \
                % (source_layer.name, target_layer.name))

        source_type = source_layer['type']
        target_type = target_layer['type']
        if (source_type != target_type):
            raise Exception("Layer types mismatch (%s -> %s)" \
                % (source_type, target_type))

        if ('params' in target_layer['config']):
            if ('W' in target_layer['config']['params']):
                source_weights = ['W']
                target_weights = ['W']
                if (source_weights.shape != target_weights.shape):
                    raise Exception("Layer weight shapes mismatch (%s -> %s)"
                        % (source_weights, target_weights))
    except Exception as e:
        return (False, str(e))
    return (True, "No error")

def import_matching_layers(source_model, target_model, load_states=True):
    source_model_layers = {l['config']['name']: l \
        for l in source_model['model']['config']['layers']}
    print("Source model has layers:", source_model_layers.keys())
    imported_layers = []
    target_model_layers = target_model.layers.layers
    for i, target_layer in enumerate(target_model_layers):
        target_layer_desc = target_layer.get_description()
        source_layer = \
            source_model_layers.get(target_layer_desc['config']['name'])
        is_match, mismatch_reason = \
            is_layer_matches(source_layer, target_layer_desc)
        if (is_match == True):
            target_layer.load_weights(source_layer, load_states)
            imported_layers.append(target_layer.name)
            print("Successfully copied layer #%d '%s'" \
                % (i, target_layer.name))
        else:
            print("Unable to copy layer #%d '%s': %s" \
                % (i, target_layer.name, mismatch_reason))
    return imported_layers